import re, os, sys

sys.path.insert(0,os.path.dirname(os.path.abspath(__file__)))

class NodeError(Exception):
  pass

class ParseError(Exception):
  pass

class XmlNode:
  """
  Represents a single element in a multi-element categorization
  of a "test.info" description. For example, a single "path" to
  a "test.info" node might look like:

    Sod/pm3/hdf5/parallel/1d

  Each element in this path would be represented by a single
  node object. The node whose name was "parallel" would have
  a sub-node with the name "1d". The "1d" node would in turn
  have no sub-nodes, but its "text" member would contain the
  information necessary to run a FLASH simulation of type "Sod",
  using paramesh3, etc.

  Nodes are typically constructed by parsing an xml file with
  this modules "parseXml()" method. For example, a text file
  that contained the following:

    <Sod>
      <pm3>
        <hdf5>
          <parallel>
            <1d>
              setupName: Sod
              numProcs: 1
              parfiles: flash.par
            </1d>
          </parallel>
        </hdf5>
      </pm3>
    </Sod>

  would produce the nodes described above when parsed.

  Calling the "Sod" node's "getXml()" method would reproduce the
  above xml source again. Thus a node's sub-nodes and/or text can
  be easily manipulated in memory and re-converted into xml.

  I considered using pickled objects and avoiding xml altogether,
  but decided that the usefulness of having the ultimate source
  of the data encoded in a human-readable form was more important
  than the degree of inefficiency introduced by parsing the xml
  file into node objects and back again.
  """

  ########################
  ##  python3 functions  ##
  ########################
  
  def __init__(self, name):
    self.name     = name
    self.parent   = None
    self.depth    = 0
    self.subNodes = []
    self.text     = []
    self.dirty = False

  def __iter__(self):
    # NB: this method returns a generator
    # object that can be iterated over
    return self.__nodeGenerator()

  def __str__(self):
    return self.name

  def __repr__(self):
    return self.name


  #########################
  ##  private functions  ##
  #########################

  def __adjustDepth(self):
    """
    Make sure the depth values of this
    node and all its children are correct
    """
    self.depth = self.parent.depth + 1
    for subNode in self.subNodes:
      subNode.__adjustDepth()

  def __adopt(self, child):
    """
    Give this node a new child-node, adjusting
    the child's parent and depth properties
    """
    # make sure the incoming child's name isn't
    # the same as an already present child's
    for subNode in self.subNodes:
      if subNode.name == child.name:
        msg = ("new node's name \"%s\" already used by sub-node of " % child.name +
               "parent \"%s\"" % self.getPathBelowRoot())
        raise NodeError(msg)

    # This node passes any text to its new child, assuming
    # the child has no text of its own. This can happen if
    # user tries an "add_node" command on a leaf node.
    if len(child.text) == 0:
      child.text = self.text[:]
    self.text = []
    self.subNodes.append(child)
    if child.parent and child in child.parent.subNodes:
      child.parent.subNodes.remove(child)
    child.parent = self
    child.__adjustDepth()

  def __copy(self):
    """
    Return a copy of this node with copies of
    all child-nodes and text
    """
    newNode = XmlNode(self.name)
    newNode.text = self.text[:]
    for subNode in self.subNodes:
      newNode.__adopt(subNode.__copy())
    return newNode

  def __nodeGenerator(self):
    """
    Returns a python3 generator object so that sub-nodes
    can be listed recursively (depth first) when a node
    is iterated over
    """
    for subNode in self.subNodes:
      yield subNode
      for subSubNode in subNode.__nodeGenerator():
        yield subSubNode


  ########################
  ##  public functions  ##
  ########################
  def smudge(self):
    self.dirty = True
    if self.parent:
      self.parent.smudge()
  
  # overwrite/add every smudged node from self onto that
  def mergeDirtyFrom(self, that):
    def seek(root, path): # will find or create
      for x in path:
        nd1 = None
        for nd in root.subNodes:
          if nd.name == x:
            nd1 = nd
            break
        if nd1 is None:
          nd1 = XmlNode(x)
          root.__adopt(nd1)
        root = nd1
      return root
    def visit(path, nd):
      if not nd.dirty: return
      nd1 = seek(self, path)
      nd1.text = list(nd.text) # shallow copy
      for sub in nd.subNodes:
        visit(path + [sub.name], sub)
    visit([], that)
  
  def add(self, newNodeName):
    """
    Create new node as child of this node with name 'newNodeName'
    """
    # check for a conflicting name
    childrenNames = [subNode.name for subNode in self.subNodes]
    if newNodeName in childrenNames:
      if self.isMasterNode():
        path = "masterNode"
      else:
        path = self.getPathBelowRoot()
      msg = ("Unable to create node with name \"%s\" " % newNodeName +
             "as sub-node of \"%s\", " % path +
             "as this node already has a sub-node of that name.")
      raise NodeError(msg)
    # else
    newNode = XmlNode(newNodeName)
    self.__adopt(newNode)
    return newNode

  def clone(self, newNodeName):
    """
    Create new sibling-node based on recursive copy
    of all sub-nodes and text.
    """
    # check for a conflicting name
    childrenNames = [subNode.name for subNode in self.parent.subNodes]
    if newNodeName in childrenNames:
      if self.parent.isMasterNode():
        parentPath = "master-node"
      else:
        parentPath = self.parent.getPathBelowRoot()
      msg = ("Unable to create node with name \"%s\" " % newNodeName +
             "as sub-node of \"%s\", " % parentPath +
             "as this node already has a sub-node of that name.")
      raise NodeError(msg)
    # else
    newNode = self.__copy()
    newNode.name = newNodeName
    self.parent.__adopt(newNode)
    return newNode

  def extend(self, newNodeName):
    """
    Create new node as child of this node with name 'newNodeName',
    but pass any of this node's current children to the new node.
    """
    newNode = XmlNode(newNodeName)
    # We don't need check for name conflicts because by
    # definition 'newNode' will be this node's only child
    for subNode in self.subNodes[:]:
      newNode.__adopt(subNode)  # this will automatically remove 'subNode'
                                # from 'self's' list of sub-nodes
    self.__adopt(newNode)
    return newNode

  def isMasterNode(self):
    """
    Returns True if this node is the master-node
    (has no parent), otherwise False
    """
    if self.parent:
      return False
    return True

  def remove(self, fullRemove):
    """
    Remove this node from its parent's list of sub-nodes.
    If 'fullRemove' is false, this node's sub-nodes will
    be adopted by its parent unless any of those sub-nodes
    has the same name as one of the parent's already-extant
    children. In that circumstance, a "NodeError" is raised.
    """
    if fullRemove:
      self.parent.subNodes.remove(self)
    else:
      # check for a conflicting name
      siblingNames = [subNode.name for subNode in self.parent.subNodes if subNode != self]
      for subNode in self.subNodes:
        if subNode.name in siblingNames:
          if self.parent.isMasterNode():
            parentPath = "master-node"
          else:
            parentPath = self.parent.getPathBelowRoot()
          msg = ("Unable to pass sub-node \"%s\" " % subNode.name +
                 "to parent-node \"%s\", " % parentPath +
                 "as parent already has a child of that name.")
          raise NodeError(msg)
      # else
      self.parent.subNodes.remove(self)
      for subNode in self.subNodes:
        self.parent.__adopt(subNode)

      # If parent had no other children besides this one,
      # it can absorb this one's text.
      if len(self.parent.subNodes) == 0:
        self.parent.text = self.text[:]

  def rename(self, newName):
    """
    Rename this node unless its parent already
    has a sub-node with name 'newName'
    """
    siblingNames = [subNode.name for subNode in self.parent.subNodes if subNode != self]
    if newName in siblingNames:
      msg = ("Unable to rename \"%s\" " % self.getPathBelowRoot() +
             "to \"%s\", as parent already has a child of that name." % newName)
      raise NodeError(msg)
    # else
    self.name = newName

  def findChild(self, path):
    """
    Return a node corresponding to 'path', where
    each element of 'path' is the name of a node.
    """
    path = os.path.normpath(path)
    if path == ".":
      return self
    # else
    currentWorkingNode = self
    pathElements = path.split(os.sep)
    while len(pathElements) > 0:
      for subNode in currentWorkingNode.subNodes:
        if subNode.name == pathElements[0]:
          currentWorkingNode = subNode
          break  # break out of the for-loop and skip the "else" clause below
      else:
        # 'path' led to a non-existant node
        return None

      del pathElements[0]

    return currentWorkingNode

  def findChildren(self, soughtName):
    """
    Return a list of nodes beneath this
    node whose names match 'soughtName'
    """
    foundNodes = []
    def __find(node):
      for subNode in node.subNodes:
        if subNode.name == soughtName:
          foundNodes.append(subNode)
        __find(subNode)

    __find(self)
    return foundNodes

  def findChildrenWithPath(self, soughtName):
    """
   Has a similar function to findChildren() but soughtName is
   a pathName like "GrandfatherNode/FatherNode/soughtNode"
    """

    foundNodes=[]
    sn = soughtName.split('/',1)        
    if len(sn) == 1: return self.findChildren(soughtName)
    
    def __subfind(node,nm):
        names = nm.split('/',1)
        for subNode in node.subNodes:
                if subNode.name == names[0]:
                        if len(names) > 1:
                                return __subfind(subNode, names[1])
                        else:
                                return subNode
                        
    def __find(node):
      for subNode in node.subNodes:
        if subNode.name == sn[0]:
          foundNodes.append(__subfind(subNode,sn[1]))
        __find(subNode)


    __find(self)
    return foundNodes

  def getPathBelowRoot(self):
    """
    Return a path from the master-node to this node
    where each element in the path corresponds to a
    node's 'name' property.
    """
    pathElements = []
    thisNode = self
    while thisNode.parent:
      pathElements.insert(0, thisNode.name)
      thisNode = thisNode.parent

    return os.sep.join(pathElements)

  def getXml(self):
    """
    Return as a list the lines of xml code that represent this node,
    its subNodes, their subNodes, etc., and any text contained in the
    leaf nodes.
    """
    xml = []
    def _getXml(xmlNode):
      tagIndent  = "  " * (xmlNode.depth - 1)
      lineIndent = "  " * xmlNode.depth
      xml.append(tagIndent + ("<%s>" % xmlNode.name))
      for subNode in xmlNode.subNodes:
        _getXml(subNode)
      for line in xmlNode.text:
        xml.append(lineIndent + line)
      xml.append(tagIndent + ("</%s>" % xmlNode.name))

    if not self.parent:
      # the all-encompassing 'masterNode' should not appear
      # in the xml. The first-level subNodes will have zero
      # left-hand indentation
      for subNode in self.subNodes:
        _getXml(subNode)
    else:
      _getXml(self)

    return xml
    
def parseXml(pathToXml):
  """
  Given a path to an xml file, create a master node and parse the xml
  file into sub-nodes of the master node. Return the master node.
  """
  with open(pathToXml) as f:
    return parseXmlString(f.read())

def parseXmlString(data):
  startTagPat = "<(?!/)(.*)>"  # matches '<' only if next char is not '/'
  endTagPat   = "</(.*)>"      # matches '<' followed by '/'

  def __parse(xmlNode, lineNum):
    while lineNum < len(xmlLines):
      thisLine = xmlLines[lineNum]
      lineNum += 1
      m = re.match(startTagPat, thisLine)
      if m:
        # 'thisLine' is a start-tag
        if len(xmlNode.text) > 0:
          if xmlNode.isMasterNode():
            msg = "Master-node contains both sub-nodes and text."
          else:
            msg = "Node \"%s\" contains both sub-nodes and text." % xmlNode.getPathBelowRoot()
          raise ParseError(msg)
        # else
        newNodeName = m.group(1)
        try:
          newNode = xmlNode.add(newNodeName)
        except NodeError as e:
          if xmlNode.isMasterNode():
            msg = "Master-node has multiple children named \"%s\"." % newNodeName
          else:
            msg = "Node \"%s\" has multiple children named \"%s\"." % (xmlNode.getPathBelowRoot(), newNodeName)
          raise ParseError(msg)
        else:
          lineNum = __parse(newNode, lineNum)
      else:
        m = re.match(endTagPat, thisLine)
        if m:
          # 'thisLine' is an end-tag
          if m.group(1) == xmlNode.name:
            return lineNum
          else:
            if xmlNode.isMasterNode():
              msg = ("closing tag \"%s\" has no corresponding opening tag." % m.group(1))
            else:
              msg = ("closing tag \"%s\" does not match opening tag \"%s\"." %
                     (m.group(1), xmlNode.getPathBelowRoot()))
            raise ParseError(msg)
        else:
          # 'thisLine' is text
          if len(xmlNode.subNodes) > 0:
            if xmlNode.isMasterNode():
              msg = "Master-node contains both sub-nodes and text."
            else:
              msg = "Node \"%s\" contains both sub-nodes and text." % xmlNode.getPathBelowRoot()
            raise ParseError(msg)
          # else
          xmlNode.text.append(thisLine)
    else:
      if xmlNode.depth > 0:
        msg = "Node \"%s\" has no closing tag." % xmlNode.getPathBelowRoot()
        raise ParseError(msg)
  
  # read in the xml text file and remove empty lines and comments
  xmlLines = data.split("\n")
  xmlLines = [xmlLine.strip() for xmlLine in xmlLines
              if (len(xmlLine.strip()) > 0 and
                  not xmlLine.startswith("#"))]

  # create the master node that will contain all the nodes in the text file
  masterNode = XmlNode("masterNode")

  # "__parse()" will read the info from 'xmlLines' into 'masterNode'
  __parse(masterNode, 0)
  return masterNode




import sys, os, time, signal, locale

sys.path.insert(0,os.path.dirname(os.path.abspath(__file__)))

CHUNKSIZE = 100               # number of bytes to read (default, for usePty)

def getProcessResults(cmd, timeout=None, usePty=False):
  cmd = cmd.strip()
  cmds = cmd.split(";")
  if len(cmds) == 1:
    return getProcessResultsInner(cmd, timeout, usePty)
  # else
  cmds = [cmd.strip() for cmd in cmds if len(cmd.strip()) > 0]
  combinedOut = ""
  combinedErr = ""
  combinedDur = ""
  for cmd in cmds:
    label = "for cmd = \"%s\":\n%s" % (cmd, "************" + "*" * len(cmd) + "\n")
    combinedOut += label
    combinedErr += label
    combinedDur += label

    out, err, dur, exitStatus = getProcessResultsInner(cmd, timeout, usePty)
    combinedOut += out + "\n"
    combinedErr += err + "\n"
    combinedDur += str(dur) + "\n"
    if exitStatus != 0:
      break

  # hack warning - if user is expecting duration to be
  # expressed as a float, as it is in the single 'cmd'
  # version, he'll be disappointed that 'combinedDur'
  # is a string.
  return (combinedOut, combinedErr, combinedDur, exitStatus)

def getProcessResultsInner(cmd, timeout=None, usePty=False):
  """
  executes 'cmd' in a separate process and returns a tuple
  containing the child's stdout and stderr together, the stderr
  separately, the duration of execution, and the process exit
  status. May use pseudo-terminals, perhaps to capture output of
  child process without buffering. Parent will abort the child if
  child fails to produce output for 'timeout' consecutive seconds.

  The command should be a string containing a valid shell command
  *without* any quoted space characters or other special characters
  embedded in the command or in arguments, since these may not be
  parsed as intended.
  DO NOT USE THIS FUNCTION TO PROCESS COMMANDS OF UNTRUSTED ORIGIN.

  Many thanks go to Jess Balint of the Chicago python3 users' group
  for invaluable help in developing an erlier version of this code,
  in which the subprocess module was not used.
  The current version uses the subprocess module unless usePty=True
  is requested. It was originally tested with Python 2.7.9 and 3.7.6.
  """
  import subprocess, select, fcntl

  global CHUNKSIZE               # refers to the global variable. This may be too weird.

  if timeout == None:
    timeout = 10 * 60  # kill process if it produces
                      # no output for 5 minutes.
                      # This is being upped to 10 for now. -PR

  useOsIO = False
  useSubprocess = not usePty

  if useSubprocess:
      p = subprocess.Popen(cmd, shell=True,
                           stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                           bufsize=0)
      (stdOutR, stdErrR, childPid) = (p.stdout, p.stderr, p.pid)
      # Python 3.5 ff. has os.set_blocking(fd,blocking); but some
      # testing has shown that explicitly setting the fd to O_NONBLOCK may not be
      # necessary for Python 3 anyway.
      fcntl.fcntl(stdOutR, fcntl.F_SETFL, os.O_NONBLOCK)
      fcntl.fcntl(stdErrR, fcntl.F_SETFL, os.O_NONBLOCK)
      CHUNKSIZE = 200 # This may not be the right number, maybe much
                      # larger is okay.  But a justification for reading
                      # only in small chunks, even though much larger
                      # ones can also work, is the following: we do not
                      # want to lose much output in case a process dies
                      # prematurely.
  else:
      useOsIO = True      # If not using the subprocess module, always
                          # use functions from os directly.
      # Use a pipe for the stderr, because it's not file-buffered,
      # and a pseudo-terminal for the stdout if usePty is True.
      (stdErrR, stdErrW) = os.pipe()
      if usePty:
          (stdOutR, stdOutW) = os.openpty()
      else:
          (stdOutR, stdOutW) = os.pipe()
      try:
          CHUNKSIZE = select.PIPE_BUF # This may not be the right number, maybe much larger is okay
      except:
          pass
      # os.fork() will return the child's pid to the parent and 'None' to the child
      childPid = os.fork()

  if childPid:  # this is the parent
    out = b""
    err = b""
    stdOutDone = False
    stdErrDone = False

    if useSubprocess:
      if useOsIO:    # In this combinations of flags, extract the
                     # underlying file descriptors from the stream
                     # objects and use them for further I/O
                     # operations; otherwise, the stream objects from
                     # the Popen object will be used directly to
                     # invoke higher-level methods for I/O operations.
        stdOutR =stdOutR.fileno()
        stdErrR =stdErrR.fileno()
    else:      # We are dealing with file descriptors, not file objects.
      # parent doesn't need this end (child does)
      os.close(stdErrW)
      os.close(stdOutW)

    startTime = time.time()
    wasAborted = False                  # will be set after SIGTERM has been sent
    wasAbortedHard = False              # will be set after SIGKILL (aka 9) has been sent
    readsAfterAbort = 0

    effectiveTimeout = timeout
    while 1:
      # wait for some results (with a possible timeout)
      (rl, wl, xl) = select.select([stdOutR, stdErrR], [], [stdOutR, stdErrR], effectiveTimeout)

      if len(rl) == 0 and not wasAborted:
        # we timed out, so kill the child (gently)
        if useSubprocess: p.send_signal(signal.SIGTERM)
        else:             os.kill(childPid, signal.SIGTERM)
        emsg = "\nProcess killed (-TERM) by FlashTest after producing no output for %s seconds.\n" % effectiveTimeout
        err += emsg.encode()
        effectiveTimeout = max(1, timeout / 20)
        wasAborted = True
        continue

      elif wasAborted and (len(rl) == 0 or readsAfterAbort >= 10):
        # we timed out, so kill the child (vigorously)
        if useSubprocess: p.kill()
        else:             os.kill(childPid, 9)
        if len(rl) == 0: emsg = "\nProcess killed (-9) by FlashTest after producing no output for another %s seconds.\n" % effectiveTimeout
        else: emsg = "\nProcess provided output %d times after being killed.\n" % readsAfterAbort
        err += emsg.encode()
        wasAbortedHard = True
        break

      # else
      if stdOutR in rl:
        try:
          # capture the output of the child. Pseudo-terminals sometimes replace
          # regular newlines '\n' with '\r\n', so replace all instances of '\r'
          # with the empty string. This behavior of pseudo-terminals makes them
          # unsafe for use when the child's output is binary data.
          if usePty:
            o = os.read(stdOutR, CHUNKSIZE).replace(b"\r",b"")
          else:
            if useOsIO:
              o = os.read(stdOutR, CHUNKSIZE)
            else:
              o = stdOutR.read(CHUNKSIZE)
          if wasAborted: readsAfterAbort += 1
        except OSError:
          # we might not be able to read the pty, so we trap the error and set
          # 'end' to True. We don't break because there might be something yet
          # to read from 'stdErrR'.
          stdOutDone = True
        else:
          try:
            len_o = len(o)
            if len_o:
              try:
                out += o
                len_out = len(out)
              except MemoryError as ex:
                o = b""
                out = b""
                out = ("***Memory Error, discarding standard output!***\n%s" % ex).encode(errors="ignore")
                stdOutDone = True
                # Something is badly wrong, so kill the child (try gently)
                if not wasAborted:
                  if useSubprocess: p.send_signal(signal.SIGTERM)
                  else:             os.kill(childPid, signal.SIGTERM)
                  out += b"\nProcess killed (-TERM) by FlashTest because a MemoryError was caught.\n"
                  err += b"\nProcess killed (-TERM) by FlashTest because a MemoryError was caught.\n"
                  effectiveTimeout = max(1, timeout / 20)
                  wasAborted = True
                else:
                  # we already tried to terminate, so kill the child (vigorously)
                  if useSubprocess: p.kill()
                  else:             os.kill(childPid, 9)
                  out += b"\nProcess killed (-9) by FlashTest because a MemoryError was caught.\n"
                  err += b"\nProcess killed (-9) by FlashTest because a MemoryError was caught.\n"
                  wasAbortedHard = True
                  break
              else:
                if len_out > 10*1024*1024:
                # Something went probably wrong, too much output to stdout.
                  if not wasAborted:
                    if useSubprocess: p.send_signal(signal.SIGTERM)
                    else:             os.kill(childPid, signal.SIGTERM)
                    out += b"\nProcess killed (-TERM) by FlashTest, more than 10 MiB of standard output.\n"
                    err += b"\nProcess killed (-TERM) by FlashTest, more than 10 MiB of standard output.\n"
                    effectiveTimeout = max(1, timeout / 20)
                    wasAborted = True
                  else:
                    # we already tried to terminate, so kill the child (vigorously)
                    if useSubprocess: p.kill()
                    else:             os.kill(childPid, 9)
                    out += b"\nProcess killed (-9) by FlashTest, more than 10 MiB of standard output.\n"
                    err += b"\nProcess killed (-9) by FlashTest, more than 10 MiB of standard output.\n"
                    wasAbortedHard = True
                    break
            else: stdOutDone = True
          except SystemError as ex:
            o = b""
            out = b""
            out = ("***System Error (maybe a memory problem?), discarding standard output!***\n%s" % ex).encode(errors="ignore")
            stdOutDone = True
            # Something is badly wrong, so kill the child (try gently)
            if not wasAborted:
              if useSubprocess: p.send_signal(signal.SIGTERM)
              else:             os.kill(childPid, signal.SIGTERM)
              out += b"\nProcess killed (-TERM) by FlashTest because a SystemError was caught.\n"
              err += b"\nProcess killed (-TERM) by FlashTest because a SystemError was caught.\n"
              effectiveTimeout = max(1, timeout / 20)
              wasAborted = True
            else:
              # we already tried to terminate, so kill the child (vigorously)
              if useSubprocess: p.kill()
              else:             os.kill(childPid, 9)
              out += b"\nProcess killed (-9) by FlashTest because a SystemError was caught.\n"
              err += b"\nProcess killed (-9) by FlashTest because a SystemError was caught.\n"
              wasAbortedHard = True
              break

      if stdErrR in rl:
        # stdErr is a pipe, not a pseudo-terminal like stdOut, so we can read
        # from the stream without a try-except block, knowing that a pipe will
        # return the empty string on end-of-file instead of raising an exception
        if useOsIO:
          e = os.read(stdErrR, CHUNKSIZE)
        else:
          e = stdErrR.read(CHUNKSIZE)
        if e:
          if wasAborted: readsAfterAbort += 1
          # put error both in regular output stream and in its own
          out += e
          err += e
        else:
          # when the child process terminates gracefully on platforms
          # other than Irix, 'stdErrR' will appear in 'rl' and a call
          # to os.read() will return the empty string (sometimes)
          stdErrDone = True

      # when the child process terminates on Irix, it shows up
      # in the 'exceptional conditions' list returned by "select"
      if len(xl) > 0:
        break

      if stdOutDone and stdErrDone:
        # break only when both stdOut and stdErr streams are empty
        break

    endTime = time.time()
    duration = endTime - startTime
    if wasAbortedHard:
      exitStatus = 9
    else:
      try:
        if useSubprocess:
          exitStatus = p.wait()
        else:
          exitStatus = os.waitpid(childPid,0)[1]
      except:  # if 'childPid' no longer exists for some reason
        if wasAborted:
          exitStatus = 9
        else:
          exitStatus = "unknown"

    if useOsIO:
      os.close(stdErrR)
      os.close(stdOutR)
    else:      
      stdErrR.close()
      stdOutR.close()
    return (out.decode(encoding='utf-8',errors="ignore"),
            err.decode(encoding=locale.getpreferredencoding(False),errors="replace"),
            duration,
            exitStatus)

  else:  # this is the child

    # child doesn't need this end (parent does)
    os.close(stdErrR)
    os.close(stdOutR)

    # make sure we're clean
    sys.stdout.flush()
    sys.stderr.flush()

    # correct the new stdout and stderr
    os.dup2(stdOutW, sys.stdout.fileno())
    os.dup2(stdErrW, sys.stderr.fileno())

    # child doesn't need these either after they've been dup'd
    os.close(stdOutW)
    os.close(stdErrW)

    # Go
    cmd = cmd.split()
    try:
      os.execvp(cmd[0], cmd)
    except Exception as e:
      print (e)  # will go to duplicate of 'stdOutW'

    sys.stdout.flush()
    sys.stderr.flush()
    os._exit(2) # make sure child exits if execvp failed.

    # end of program. Nothing should happen after
    # 'exec' if the child has executed correctly.



import os,sys

sys.path.insert(0,os.path.dirname(os.path.abspath(__file__)))

from getProcessResults import getProcessResults

class ArchiveError(Exception):
  pass

class Archive:

  def __init__(self, log, masterDict):
    self.log = log

    # these are all guaranteed to exist in 'masterDict' by FlashTest.py
    self.pathToFlashTest     = masterDict["pathToFlashTest"]
    self.pathToSiteDir       = masterDict["pathToSiteDir"]
    self.pathToInvocationDir = masterDict["pathToInvocationDir"]
    self.pathToLocalArchive  = masterDict["pathToLocalArchive"]
    self.mainArchiveHost     = masterDict["mainArchiveHost"]
    self.pathToMainArchive   = masterDict["pathToMainArchive"]
    self.viewArchiveHost     = masterDict["viewArchiveHost"]
    self.pathToViewArchive   = masterDict["pathToViewArchive"]
    self.logArchiveHost      = masterDict["logArchiveHost"]
    self.pathToLogArchive    = masterDict["pathToLogArchive"]

    # these two might or might not exist in 'masterDict'
    self.pathToMainArchiveGnuTar = masterDict.get("pathToMainArchiveGnuTar", "tar")
    self.pathToGnuTar            = masterDict.get("pathToGnuTar", "tar")

  def confirmInLocalArchive(self, pathToLocalFileOrDir):
    """
    Check for existence of 'pathToLocalFileOrDir' in local archive. If
    not there, extract it from requisite .tar.gz file in main archive.
    Return True on success or raise ArchiveError on failure.
    """
    pathToLocalFileOrDir = os.path.normpath(pathToLocalFileOrDir)

    log                = self.log
    pathToLocalArchive = self.pathToLocalArchive

    if not os.path.isdir(pathToLocalArchive):
      raise ArchiveError("\"%s\" does not exist or is not a directory." % pathToLocalArchive)

    if os.path.isabs(pathToLocalFileOrDir):
      if not pathToLocalFileOrDir.startswith(pathToLocalArchive):
        raise ArchiveError("Path \"%s\" should be an absolute path to a file beneath" % pathToLocalFileOrDir +
                           "local archive \"%s\"" % pathToLocalArchive +
                           "or a relative path which will be interpreted as a path" +
                           "to a file beneath \"%s\"" % pathToLocalArchive)
    else:
      pathToLocalFileOrDir = os.path.join(pathToLocalArchive, pathToLocalFileOrDir)

    # ideally, 'pathToLocalFileOrDir' already exists, i.e. the
    # requested file or directory is already in the local archive
    if os.path.exists(pathToLocalFileOrDir):
      return True

    # reconstruct the relative path from the local archive to the file or directory
    shortPathToLocalFileOrDir = pathToLocalFileOrDir[len(pathToLocalArchive):].strip(os.sep)

    # try to find the necessary file or directory in the main archive
    log.info("\"%s\" not found in local archive.\n" % shortPathToLocalFileOrDir +
             "Searching the main archive...")

    pathToMainArchive = self.pathToMainArchive
    if not pathToMainArchive:
      raise ArchiveError("No path to main archive provided.")

    if not os.path.isabs(pathToMainArchive):
      raise ArchiveError("Path to main archive \"%s\" must be an absolute path." % pathToMainArchive)

    # 'shortPathToLocalFileOrDir' becomes even shorter as
    # the site and the invocation are chopped off the front
    site, shortPathToLocalFileOrDir  = shortPathToLocalFileOrDir.split(os.sep,1)
    invoc, shortPathToLocalFileOrDir = shortPathToLocalFileOrDir.split(os.sep,1)
    pathToArchivedTarFile            = os.path.join(pathToMainArchive, site, invoc) + ".tar.gz"
    mainArchiveHost                  = self.mainArchiveHost

    # copy the archived tarfile into the local archive if possible
    if mainArchiveHost:  # archive is on a remote computer

      # test for existence of archived tar file
      cmd = "ssh %s test -e %s" % (mainArchiveHost, pathToArchivedTarFile)
      out, err, duration, exitStatus = getProcessResults(cmd)

      if err:
        raise ArchiveError("Unable to connect to %s\n%s" % (mainArchiveHost,
                                                            err))
      # else
      if exitStatus != 0:
        raise ArchiveError("File \"%s\" not found on \"%s\"" % (pathToArchivedTarFile,
                                                                mainArchiveHost))

      # else
      pathToMainArchiveGnuTar     = self.pathToMainArchiveGnuTar
      pathToArchivedTarFileParent = os.path.dirname(pathToArchivedTarFile)

      cmd = "ssh %s %s xvzf %s -C %s %s" % (mainArchiveHost,
                                           pathToMainArchiveGnuTar,
                                           pathToArchivedTarFile,
                                           pathToArchivedTarFileParent,
                                           os.path.join(invoc, shortPathToLocalFileOrDir))
      out, err, duration, exitStatus = getProcessResults(cmd)

      if err:
        raise ArchiveError("Unable to extract \"%s\" from \"%s:%s\"\n%s" %
                           (os.path.join(invoc, shortPathToLocalFileOrDir),
                            mainArchiveHost,
                            pathToArchivedTarFile,
                            err))
      # else
      if exitStatus != 0:
        raise ArchiveError("Exit status %s indicates error extracting \"%s\" from \"%s:%s\"" %
                           (exitStatus,
                            os.path.join(invoc, shortPathToLocalFileOrDir),
                            mainArchiveHost, pathToArchivedTarFile))

      # else prepare to bring remote copy over to local archive.
      # Make necessary directories if they don't already exist
      pathToLocalFileOrDirParent = os.path.dirname(pathToLocalFileOrDir)

      if not os.path.isdir(pathToLocalFileOrDirParent):
        try:
          os.makedirs(pathToLocalFileOrDirParent)
        except Exception as e:
          raise ArchiveError("Unable to create directory\n%s" % e)

      # else
      cmd = "scp -r %s:%s %s" % (mainArchiveHost,
                                 os.path.join(pathToArchivedTarFileParent, invoc, shortPathToLocalFileOrDir),
                                 pathToLocalFileOrDir)
      out, err, duration, exitStatus = getProcessResults(cmd)

      if err:
        raise ArchiveError("Unable to copy \"%s:%s\" to local archive.\n%s" %
                           (mainArchiveHost,
                            pathToArchivedTarFile,
                            err))
      # else
      if exitStatus != 0:
        raise ArchiveError("Exit status %s indicates error copying \"%s:%s\" to local archive." %
                           (exitStatus,
                            mainArchiveHost, pathToArchivedTarFile))

      # else delete extracted copy of file on remote host
      cmd = "ssh %s rm -rf %s" % (mainArchiveHost,
                                  os.path.join(pathToArchivedTarFileParent, invoc))
      out, err, duration, exitStatus = getProcessResults(cmd)

      if err:
        log.warn("Unable to delete remote copy of extracted archive \"%s:%s\"" %
                 (mainArchiveHost,
                  os.path.join(pathToArchivedTarFileParent, invoc),
                  err))
      # else
      if exitStatus != 0:
        log.warn("Exit status %s indicates error deleting remote copy of extracted archive \"%s:%s\"" %
                 (exitStatus,
                  mainArchiveHost,
                  os.path.join(pathToArchivedTarFileParent, invoc)))
        
    else:  #  archive is on local computer

      # test for existence of archived tar file
      if not os.path.isfile(pathToArchivedTarFile):
        raise ArchiveError("Unable to find file \"%s\"" % pathToArchivedTarFile)

      # else
      pathToGnuTar                = self.pathToGnuTar
      pathToArchivedTarFileParent = os.path.dirname(pathToArchivedTarFile)

      cmd = "%s xvzf %s -C %s %s" % (pathToGnuTar,
                                     pathToArchivedTarFile,
                                     pathToArchivedTarFileParent,
                                     os.path.join(invoc, shortPathToLocalFileOrDir))
      out, err, duration, exitStatus = getProcessResults(cmd)

      if err:
        raise ArchiveError("Unable to extract \"%s\" from \"%s\"\n%s" %
                           (os.path.join(invoc, shortPathToLocalFileOrDir),
                            pathToArchivedTarFile,
                            err))
      # else
      if exitStatus != 0:
        raise ArchiveError("Exit status %s indicates error extracting \"%s\" from \"%s\"" %
                           (exitStatus,
                            os.path.join(invoc, shortPathToLocalFileOrDir),
                            pathToArchivedTarFile))

      # else prepare to bring remote copy over to local archive.
      # Make necessary directories if they don't already exist
      pathToLocalFileOrDirParent = os.path.dirname(pathToLocalFileOrDir)
      if not os.path.isdir(pathToLocalFileOrDirParent):
        try:
          os.makedirs(pathToLocalFileOrDirParent)
        except Exception as e:
          raise ArchiveError("Unable to create directory\n%s" % e)


      # else
      cmd = "cp -r %s %s" % (os.path.join(pathToArchivedTarFileParent, invoc, shortPathToLocalFileOrDir),
                             pathToLocalFileOrDir)
      out, err, duration, exitStatus = getProcessResults(cmd)

      if err:
        raise ArchiveError("Unable to copy \"%s\" to local archive.\n%s" %
                           (pathToArchivedTarFile,
                            err))
      # else
      if exitStatus != 0:
        raise ArchiveError("Exit status %s indicates error copying \"%s\" to local archive." %
                           (exitStatus,
                            pathToArchivedTarFile))

      # else delete extracted copy of file
      cmd = "rm -rf %s" % (os.path.join(pathToArchivedTarFileParent, invoc))
      out, err, duration, exitStatus = getProcessResults(cmd)

      if err:
        log.warn("Unable to delete remote copy of extracted archive \"%s\"" %
                 (os.path.join(pathToArchivedTarFileParent, invoc),
                  err))
      # else
      if exitStatus != 0:
        log.warn("Exit status %s indicates error deleting remote copy of extracted archive \"%s\"" %
                 (exitStatus,
                  os.path.join(pathToArchivedTarFileParent, invoc)))

    # confirm that file now exists in local archive
    if os.path.exists(pathToLocalFileOrDir):
      log.stp("Successfully transfered \"%s\" to \"%s\"" % (shortPathToLocalFileOrDir, pathToLocalArchive))
      if os.stat(pathToLocalFileOrDir)[6] == 0:
        log.warn("\"%s\" is zero bytes" % pathToLocalFileOrDir)
      return True

    #else
    raise ArchiveError("Was unable to transfer \"%s\" to \"%s\"" %
                       (shortPathToLocalFileOrDir, pathToLocalArchive))


  def makeTarFile(self):
    """
    Create a .tar.gz of this invocation's data
    """
    pathToGnuTar        = self.pathToGnuTar
    pathToInvocationDir = self.pathToInvocationDir
    pathToTarFile       = pathToInvocationDir + ".tar.gz"

    pathToInvocationDirParent, invocationDir = os.path.split(pathToInvocationDir)

    # -C means tar will first cd into 'pathToInvocationDirParent' before creating
    # the archive, so that files will not be archived with absolute path names
    cmd = "%s cvzf %s -C %s %s" % (pathToGnuTar,
                                   pathToTarFile,
                                   pathToInvocationDirParent,
                                   invocationDir)
    out, err, duration, exitStatus = getProcessResults(cmd)

    if err:
      if os.path.exists(pathToTarFile):
        os.remove(pathToTarFile)
      raise ArchiveError("Error creating archive \"%s\"\n%s" %
                         (pathToTarFile,
                          err))
    # else
    if exitStatus != 0:
      if os.path.exists(pathToTarFile):
        os.remove(pathToTarFile)
      raise ArchiveError("Exit status %s indicates error creating archive \"%s\"" %
                         (exitStatus, pathToTarFile))

    # else
    if not os.path.isfile(pathToTarFile):
      raise ArchiveError("Error creating archive \"%s\"" %
                         pathToTarFile)

    # else
    return pathToTarFile


  def sendTarFileToMainArchive(self, pathToTarFile):
    """
    copy tarfile indicated by 'pathToTarFile' to main
    archive, which might reside on a remote host
    """
    log               = self.log
    mainArchiveHost   = self.mainArchiveHost
    pathToMainArchive = self.pathToMainArchive

    if not pathToMainArchive:
      raise ArchiveError("No path to main archive provided.\n" +
                         "Unable to archive \"%s\"" %
                         pathToTarFile)
    # else
    pathToMainArchiveSiteDir = os.path.join(pathToMainArchive, os.path.basename(self.pathToSiteDir))

    if mainArchiveHost:  # main archive is on a remote computer
      # test if a directory with this site name exists on the main archive host
      cmd = "ssh %s test -d %s" % (mainArchiveHost, pathToMainArchiveSiteDir)
      out, err, duration, exitStatus = getProcessResults(cmd)

      if err:
        raise ArchiveError("Unable to connect to %s\n%s" % (mainArchiveHost,
                                                            err))
      # else
      if exitStatus == 0:
        # The directory does exist. Does a tarfile with this name already exist too?
        tarFile = os.path.basename(pathToTarFile)
        cmd = "ssh %s test -e %s" % (mainArchiveHost, os.path.join(pathToMainArchiveSiteDir, tarFile))
        out, err, duration, exitStatus = getProcessResults(cmd)

        if err:
          raise ArchiveError("Unable to connect to %s\n%s" % (mainArchiveHost,
                                                              err))
        # else
        if exitStatus == 0:
          # a tarfile with this name already exists. We will not overwrite
          raise ArchiveError("A file \"%s\" already exists on \"%s\"\n" %
                             (os.path.join(pathToMainArchiveSiteDir, tarFile), mainArchiveHost) +
                             "Aborting copy to main archive")
      else:
        # directory doesn't exist, so try to create it
        cmd = "ssh %s mkdir %s" % (mainArchiveHost, pathToMainArchiveSiteDir)
        out, err, duration, exitStatus = getProcessResults(cmd)

        if err:
          raise ArchiveError("Error creating directory \"%s\" on host \"%s\"\n%s" %
                             (pathToMainArchiveSiteDir,
                             mainArchiveHost,
                             err))
        # else
        if exitStatus != 0:
          raise ArchiveError("Exit status %s indicates error creating directory \"%s\" on host \"%s\"" %
                             (exitStatus, pathToMainArchiveSiteDir, mainArchiveHost))

        # else we assume the directory was created

      # copy the tarfile into the remote archive
      #cmd = "scp %s %s:%s" % (pathToTarFile, mainArchiveHost, pathToMainArchiveSiteDir)
      cmd = "rsync -ac %s %s:%s" % (pathToTarFile, mainArchiveHost, pathToMainArchiveSiteDir)
      out, err, duration, exitStatus = getProcessResults(cmd)

      if err:
        log.warn(out)
        raise ArchiveError("Error transfering \"%s\" to \"%s:%s\"\n%s" %
                           (pathToTarFile,
                            mainArchiveHost,
                            pathToMainArchiveSiteDir,
                            err))
      # else
      if exitStatus != 0:
        raise ArchiveError("Exit status %s indicates error transfering \"%s\" to \"%s:%s\"" %
                           (exitStatus, pathToTarFile, mainArchiveHost, pathToMainArchiveSiteDir))

      # else we assume archive was successfully copied

    else:  # main archive is on local computer
      # test if a directory with this site name exists
      if not os.path.isdir(pathToMainArchiveSiteDir):
        try:
          os.mkdir(pathToMainArchiveSiteDir)
        except Exception as e:
          raise ArchiveError(e)

      cmd = "cp %s %s" % (pathToTarFile,
                          pathToMainArchiveSiteDir)
      out, err, duration, exitStatus = getProcessResults(cmd)

      if err:
        raise ArchiveError("Error copying \"%s\" to \"%s\"\n%s" %
                           (pathToTarFile,
                            pathToMainArchiveSiteDir,
                            err))
      # else
      if exitStatus != 0:
        raise ArchiveError("Exit status %s indicates error copying \"%s\" to \"%s\"" %
                           (exitStatus, pathToTarFile, pathToMainArchiveSiteDir))

      # else we assume archive was successfully copied
    return True





  def sendToViewArchive(self,incremental=False):
    """
    Recursively copy this invocation's data to the view archive,
    which might reside on a remote host. This is very similar to
    "sendToMainArchive()", but doesn't create a tarball
    MODIFICATION: Add incremental parameter so that this function can be 
    called after each specific test completes i.e multiple times in a test run
    """
    viewArchiveHost     = self.viewArchiveHost
    pathToViewArchive   = self.pathToViewArchive
    pathToInvocationDir = self.pathToInvocationDir
    pathToTarFile       = pathToInvocationDir + ".tar.gz"
    invocationDir       = os.path.basename(pathToInvocationDir)

    if not pathToViewArchive:
      raise ArchiveError("No path to view archive provided.\n" +
                         "Unable to archive data from \"%s\"" % self.pathToInvocationDir)
    # else
    pathToViewArchiveSiteDir = os.path.join(pathToViewArchive, os.path.basename(self.pathToSiteDir))

    if viewArchiveHost:  # view archive is on a remote computer
      # test if a directory with this site name exists on the view archive host
      cmd = "ssh %s test -d %s" % (viewArchiveHost, pathToViewArchiveSiteDir)
      out, err, duration, exitStatus = getProcessResults(cmd)

      if err:
        raise ArchiveError("Unable to connect to %s\n%s" % (viewArchiveHost,
                                                            err))
      # else
      if exitStatus == 0:
        # The directory does exist. Does a subdirectory with the name invocationDir already exist, too?
        cmd = "ssh %s test -d %s" % (viewArchiveHost, os.path.join(pathToViewArchiveSiteDir, invocationDir))
        out, err, duration, exitStatus = getProcessResults(cmd)

        if err:
          raise ArchiveError("Unable to connect to %s\n%s" % (viewArchiveHost,
                                                              err))
        # else
        if exitStatus == 0:
          if incremental == False:      
                # a directory with this name already exists. We will not overwrite
                raise ArchiveError("A directory \"%s\" already exists on \"%s\"\n" %
                             (os.path.join(pathToViewArchiveSiteDir, invocationDir), viewArchiveHost) +
                             "Aborting copy to view archive.")
      else:
        # directory doesn't exist, so try to create it
        cmd = "ssh %s mkdir %s" % (viewArchiveHost, pathToViewArchiveSiteDir)
        out, err, duration, exitStatus = getProcessResults(cmd)

        if err:
          raise ArchiveError("Error creating directory \"%s\" on host \"%s\"\n%s" %
                             (pathToViewArchiveSiteDir,
                              viewArchiveHost,
                              err))
        # else
        if exitStatus != 0:
          raise ArchiveError("Exit status %s indicates error creating directory \"%s\" on host \"%s\"" %
                             (exitStatus, pathToViewArchiveSiteDir, viewArchiveHost))

        # else we assume the directory was created


      # recursively copy the data into the remote view archive 
      cmd = "rsync -rcltpzH --delete %s %s:%s" % (pathToInvocationDir, viewArchiveHost, pathToViewArchiveSiteDir)
      out, err, duration, exitStatus = getProcessResults(cmd, 1800)

      if err:
        raise ArchiveError("Error transferring \"%s\" to \"%s:%s\"\n%s" %
                           (pathToInvocationDir,
                            viewArchiveHost,
                            pathToViewArchiveSiteDir,
                            err))
      # else
      if exitStatus != 0:
        raise ArchiveError("Exit status %s indicates error transferring \"%s\" to \"%s:%s\"" %
                           (exitStatus, pathToInvocationDir, viewArchiveHost, pathToViewArchiveSiteDir))

      # else we assume archive was successfully copied.

      # Remove .lock file if one was copied, so that FlashTestView will NOW show the directory.
      cmd = "ssh %s rm %s" % (viewArchiveHost, os.path.join(pathToViewArchiveSiteDir,invocationDir,".lock"))
      out, err, duration, exitStatus = getProcessResults(cmd)
      if err:
        log.warn("Error from \"%s\": %s" % (cmd, err))
      if exitStatus != 0:
        if len(out):
          log.warn("Exit status %s indicates error from \"%s\", %s" % (exitStatus, cmd, out))
        else:
          log.warn("Exit status %s indicates error from \"%s\"." % (exitStatus, cmd))

      # Update last access time on view archive so FlashTestView will
      # refresh page on next browser hit. However, this doesn't actually
      # have the desired effect; touching alone is not sufficient to trigger
      # updating by FlashTestView, apparently.
      cmd = "ssh %s touch %s" % (viewArchiveHost, pathToViewArchiveSiteDir)
      getProcessResults(cmd)  # never mind if it worked or not - we tried our best

    else:  # view archive is on local computer
      # test if a directory with this site name exists
      if not os.path.isdir(pathToViewArchiveSiteDir):
        try:
          os.mkdir(pathToViewArchiveSiteDir)
        except Exception as e:
          raise ArchiveError(e)

      #cmd = "cp -r %s %s" % (pathToInvocationDir, pathToViewArchiveSiteDir)
      cmd = "rsync -rcxltH --safe-links --chmod=o+rX %s %s" % (pathToInvocationDir, pathToViewArchiveSiteDir)
      out, err, duration, exitStatus = getProcessResults(cmd)

      if err:
        raise ArchiveError("Error copying \"%s\" to \"%s\"\n%s" %
                           (pathToTarFile,
                            pathToViewArchiveSiteDir,
                            err))
      # else
      if exitStatus != 0:
        raise ArchiveError("Exit status %s indicates error copying \"%s\" to \"%s\"" %
                           (exitStatus, pathToTarFile, pathToViewArchiveSiteDir))

      # else we assume archive was successfully copied.

      # Remove .lock file if one was copied, so that FlashTestView will NOW show the directory.
      try:
        os.remove(os.path.join(pathToViewArchiveSiteDir, invocationDir, ".lock"))
      except Exception:
        pass # oh well, we did our best

      # Update last access time on view archive so FlashTestView will
      # refresh page on next browser hit.
      try:
        os.utime(pathToViewArchive, None)
      except Exception:
        pass # oh well, we did our best

    return True




  def sendToLogArchive(self,incremental=False):
    """
    Recursively copy this invocation's log files and some similar files
    to the log  archive, which might reside on a remote host. This is very
    similar to "sendToViewArchive()".
    Like that one it has an "incremental" parameter so that this function can be 
    called after each specific test completes, i.e., multiple times in a test run,
    although that may not be very useful.
    """
    log                = self.log
    logArchiveHost     = self.logArchiveHost
    pathToLogArchive   = self.pathToLogArchive
    pathToInvocationDir = self.pathToInvocationDir
    pathToTarFile       = pathToInvocationDir + ".tar.gz"

    if not pathToLogArchive:
      raise ArchiveError("No path to log archive provided.\n" +
                         "Unable to archive data from \"%s\"" % self.pathToInvocationDir)
    # else
    pathToLogArchiveSiteDir = os.path.join(pathToLogArchive, os.path.basename(self.pathToSiteDir))
    pathToLogArchiveInvocationDir = os.path.join(pathToLogArchiveSiteDir, os.path.basename(self.pathToInvocationDir))

    if logArchiveHost:  # log archive is on a remote computer
      # test if a directory with this site name exists on the log archive host
      cmd = "ssh %s test -d %s" % (logArchiveHost, pathToLogArchiveSiteDir)
      out, err, duration, exitStatus = getProcessResults(cmd)

      if err:
        raise ArchiveError("Unable to connect to %s\n%s" % (logArchiveHost,
                                                            err))
      # else
      if exitStatus == 0:
        # The directory does exist. Does a directory with this name already exist too?
        invocationDir = os.path.basename(pathToInvocationDir)
        cmd = "ssh %s test -d %s" % (logArchiveHost, os.path.join(pathToLogArchiveSiteDir, invocationDir))
        out, err, duration, exitStatus = getProcessResults(cmd)

        if err:
          raise ArchiveError("Unable to connect to %s\n%s" % (logArchiveHost,
                                                              err))
        # else
        if exitStatus == 0:
          if False: # was if incremental == False:      
                # a directory with this name already exists. We will not overwrite
                raise ArchiveError("A directory \"%s\" already exists on \"%s\"\n" %
                             (os.path.join(pathToLogArchiveSiteDir, invocationDir), logArchiveHost) +
                             "Aborting copy to log archive.")
      else:
        # directory doesn't exist, so try to create it
        cmd = "ssh %s mkdir -p %s" % (logArchiveHost, pathToLogArchiveInvocationDir)
        out, err, duration, exitStatus = getProcessResults(cmd)

        if err:
          raise ArchiveError("Error creating directory \"%s\" on host \"%s\"\n%s" %
                             (pathToLogArchiveInvocationDir,
                              logArchiveHost,
                              err))
        # else
        if exitStatus != 0:
          raise ArchiveError("Exit status %s indicates error creating directory \"%s\" on host \"%s\"" %
                             (exitStatus, pathToLogArchiveInvocationDir, logArchiveHost))

        # else we assume the directory was created


      # recursively copy the data into the remote log archive
##      cmd = "rsync -rcltpzH --delete %s %s:%s" % (pathToInvocationDir, logArchiveHost, pathToLogArchiveSiteDir)
      cmd = "rsync -mrcltpz -f +_.lock -f +_*.log -f +_*_summary* -f +_*_time -f +_test_output -f +_flash_output -f +_errors -f +_*/ -f -_* %s/ %s:%s" \
              % (pathToInvocationDir, logArchiveHost, pathToLogArchiveInvocationDir)
      #log.info("About to run command:\n%s" % cmd)
      out, err, duration, exitStatus = getProcessResults(cmd, 1800)

      if err:
        raise ArchiveError("Error transferring log files from \"%s/\" to \"%s:%s\"\n%s" %
                           (pathToInvocationDir,
                            logArchiveHost,
                            pathToLogArchiveInvocationDir,
                            err))
      # else
      if exitStatus != 0:
        raise ArchiveError("Exit status %s indicates error transferring log files from \"%s/\" to \"%s:%s\"" %
                           (exitStatus, pathToInvocationDir, logArchiveHost, pathToLogArchiveInvocationDir))

      # else we assume archive was successfully copied.

      # Remove .lock file if one was copied.
      cmd = "ssh %s rm %s" % (logArchiveHost, os.path.join(pathToLogArchiveSiteDir,os.path.basename(pathToInvocationDir),".lock"))
      out, err, duration, exitStatus = getProcessResults(cmd)
      if err:
        log.warn("Error from \"%s\": %s" % (cmd, err))
      if exitStatus != 0:
        if len(out):
          log.warn("Exit status %s indicates error from \"%s\", %s" % (exitStatus, cmd, out))
        else:
          log.warn("Exit status %s indicates error from \"%s\"." % (exitStatus, cmd))

      # Update last access time on log archive.
      cmd = "ssh %s touch %s" % (logArchiveHost, pathToLogArchiveSiteDir)
      getProcessResults(cmd)  # never mind if it worked or not - we tried our best

    else:  # log archive is on local computer
      # test if a directory with this site name exists
      if not os.path.isdir(pathToLogArchiveSiteDir):
        try:
          os.mkdir(pathToLogArchiveSiteDir)
        except Exception as e:
          raise ArchiveError(e)

      #cmd = "cp -r %s/*.log %s" % (pathToInvocationDir, pathToLogArchiveInvocationDir)
      cmd = "rsync -mrc --chmod=g+w -f +_.lock -f +_*.log -f +_*_summary* -f +_*_time -f +_test_output -f +_flash_output -f +_errors -f +_*/ -f -_* %s %s" \
              % (pathToInvocationDir, pathToLogArchiveSiteDir)
      log.info("About to run command:\n%s" % cmd)
      oldumask = os.umask(0o077)
      os.umask(oldumask & 0o757) # temporarily, do not mask the group write bit.
      out, err, duration, exitStatus = getProcessResults(cmd,usePty=False)
      os.umask(oldumask)

      if err:
        raise ArchiveError("Error copying log file(s) from \"%s\" to \"%s\"\n%s" %
                           (pathToInvocationDir,
                            pathToLogArchiveInvocationDir,
                            err))
      # else
      if exitStatus != 0:
        raise ArchiveError("Exit status %s indicates error copying log file(s) from \"%s\" to \"%s\"" %
                           (exitStatus, pathToInvocationDir, pathToLogArchiveInvocationDir))

      # else we assume archive was successfully copied.

      # Remove .lock file if one was copied.
      try:
        os.remove(os.path.join(pathToLogArchiveInvocationDir, ".lock"))
      except Exception:
        pass # oh well, we did our best

      # Update last access time on log archive.
      try:
        os.utime(pathToLogArchive, None)
      except Exception:
        pass # oh well, we did our best

    return True







import sys, os
from time import strftime, localtime

class Logfile:

  indentLen = 9

  def __init__(self, outdir, logfileName, verbose=False):
    self.outdir = outdir
    self.logfile = os.path.join(self.outdir, logfileName)
    self.verbose = verbose

    if os.path.exists(self.logfile):
      # make sure we start with a blank logfile
      os.remove(self.logfile)

  def info(self, msg, indent=True):
    self.__msg(msg, False, indent)

  def stp(self, msg):
    self.__msg(msg, True)

  def note(self, msg):
    self.__msg("NOTE!    " + msg)

  def warn(self, msg):
    self.__msg("WARNING: " + msg)

  def err(self, msg):
    self.__msg("ERROR:   " + msg)

  def brk(self):
    self.__msg("-"*80)

  def __msg(self, msg, stp=False, indent=False):
    """
    Write 'msg' to logfile after prepending timestamp (if 'stp' is True)
    """
    if stp:
      msg = "%s %s" % (strftime("%H:%M:%S", localtime()), msg)

    lines = msg.strip().split("\n")
    for i in range(len(lines)):
      if i==0 and indent==False:
        continue
      else:
        lines[i] = " " * self.indentLen + lines[i]

    msg = "\n".join(lines)

    if self.verbose:
      # stdout is redirected to "test_output" in the
      # testing phase, so temporarily override this
      stdout = sys.stdout
      sys.stdout = sys.__stdout__
      print(msg)
      sys.stdout = stdout

    if not msg.endswith("\n"):
      msg = msg + "\n"

    open(self.logfile, "a").write(msg)



"""FlashXTest library to interface with backend.FlashTest"""

import os, sys, subprocess

from .. import backend
from .. import lib


def flashTest(mainDict, jobList):
    """
    Run flashTest.py from backend/FlashTest

    Arguments:
    ----------
    Arguments:
    mainDict  : Main dictionary
    jobList   : List of jobs
    """
    # remove site from jobList
    jobList = [job.replace(f'{mainDict["flashSite"]}/', "") for job in jobList]

    # Create output directory for TestResults if it does not exist
    subprocess.run("mkdir -pv {0}".format(mainDict["pathToOutdir"]), shell=True)

    # Create archive directory if it does not exist
    subprocess.run("mkdir -pv {0}".format(mainDict["pathToLocalArchive"]), shell=True)

    # Create baseLine directory if it does not exist
    subprocess.run("mkdir -pv {0}".format(mainDict["baselineDir"]), shell=True)

    optString = __getOptString(mainDict)

    # run backend/FlashTest/flashTest.py with desired configuration
    #
    testProcess = subprocess.run(
        "python3 {0}/FlashTest/flashTest.py \
                                          {1} \
                                          {2}".format(
            os.path.dirname(backend.__file__), optString, " ".join(jobList)
        ),
        shell=True,
    )

    os.environ["EXITSTATUS"] = str(testProcess.returncode)
    os.environ["FLASH_BASE"] = mainDict["pathToFlash"]
    os.environ["RESULTS_DIR"] = (
        mainDict["pathToOutdir"] + os.sep + mainDict["flashSite"]
    )

    # try:
    checkProcess = subprocess.run(
        "bash $FLASHTEST_BASE/error.sh", shell=True, check=True
    )

    print(lib.colors.OKGREEN + "[FlashXTest] SUCCESS")

    # except checkProcess.CalledProcessError as e:
    #    #print(lib.colors.FAIL + f"{e.output}")
    #    print(e.output)


def buildSFOCU(mainDict):
    """
    Build SFOCU (Serial Flash Output Comparison Utility)

    Arguments:
    ----------
    mainDict: Dictionary from Config file
    """
    # Cache value of current directory
    workingDir = os.getenv("PWD")

    # Build brand new version of sfocu
    # cd into sfocu directory and compile a new
    # version
    os.chdir("{0}/tools/sfocu".format(mainDict["pathToFlash"]))
    subprocess.run(
        "make SITE={0} NO_NCDF=True sfocu clean".format(mainDict["flashSite"]),
        shell=True,
    )
    subprocess.run(
        "make SITE={0} NO_NCDF=True sfocu".format(mainDict["flashSite"]), shell=True
    )

    # Append SFOCU path to PATH
    os.environ["PATH"] += os.path.pathsep + os.getcwd()

    # cd back into workingDir
    os.chdir(workingDir)


def __getOptString(mainDict):
    """
    Argument
    --------

    mainDict: Dictionary with configuration values
    """
    optDict = {
        "pathToFlash": "-z",
        "pathToInfo": "-i",
        "pathToOutdir": "-o",
        "pathToConfig": "-c",
        "flashSite": "-s",
    }

    optString = "-v -L "

    for option in optDict:
        if option in mainDict:
            optString = optString + "{0} {1} ".format(optDict[option], mainDict[option])

    return optString



import io
import os, re, subprocess, time, datetime, sys
import secondsToHuman
import shutil

sys.path.insert(0,os.path.dirname(os.path.abspath(__file__)))

from getProcessResults import getProcessResults
from strategyTemplates import *
from xmlNode import *
import flashTestParser as parser

######################################
##  PRE-PROCESS, SETUP, COMPILE, &  ##
##  EXECUTE FOR ALL FLASH PROBLEMS  ##
######################################

def __abort(msg):
  sys.stderr.write(msg)
  sys.exit(1)

def _write_to_file(filename, mode, text, encoding=None, errors=None):
    """
    Helper function for writing a text in memory to a file.

    The file of the given filename is opened with the given mode and, if
    appropriate, with the given encoding and errors parameters; text is
    written to it; then the file is closed.


    This helper exists so that this file can be used in both Python2 (as
    long as it is not too ancient) and Python3.  In the Python3 case,
    the text should always be given as a string; the function may also
    work for a binary buffer, but that has not been tested.  In the
    Python2 case, this routine should work for text being either a
    unicode string or a regular string; in the latter case, the encoding
    and error parameters are ignored.

    There are many places in the code below that follow the pattern

       [io.]open( <filename>, <other arguments> ).write( <something> )

    were this function could be used instead of the existing code.  That
    might even make the code more readable.  Currently, the instances
    where the pattern has been replaced by a call to this function are
    mostly those where version compatibility problems were either
    observed or are suspected.
    """
    try:                        # u"" may be unrecognized in early python33 versions
        isU = (type(text) == type(u""))
    except SyntaxError:
        isU = (type(text) == type("")) and (type(text) != type(b""))
    if isU:
        with io.open(filename, mode, encoding=encoding, errors=errors) as f:
            f.write(text)
    else: # if type(text) == type(b""):
        with open(filename, mode) as f:
            f.write(text)

def splitHostAndPath(path):
  if path.count(':') > 0:
    return path.split(':', 1)
  else:
    return None, path
def joinHostAndPath(host,path):
  if host:
    return host + ":" + path
  else:
    return path

def pullfile(path,log):
  if path.count(':') > 0:
    host, path = path.split(':', 1)
  else:
    return path
  tmp = 'delete_me-%x' % hash(path)
  log.stp("Pulling %s:%s to %s" % (host, path, tmp))
  #print 'pulling %s: %s' % (tmp,path)
  out, err, duration, exitStatus = getProcessResults("scp %s:%s %s" % (host, path, tmp))
  if err:       #__abort("Unable to retrieve \"%s:%s\"\n%s" % (host, path, err))
    log.err("Unable to retrieve \"%s:%s\"\n%s" % (host, path, err))
    return None

  if exitStatus != 0:      #__abort("Exit status %s indicates error retrieving \"%s:%s\"" % (exitStatus, host, path))
    log.err("Exit status %s indicates error retrieving \"%s:%s\"" % (exitStatus, host, path))
    return None

  if not os.path.isfile(tmp): #__abort("File \"%s\" not found" % tmp)
    log.err("File \"%s\" not found" % tmp)
    return None

  f = open(tmp+'_mtime', 'w')
  f.write(str(os.path.getmtime(tmp)))
  f.close()
  return os.path.join(os.getcwd(),tmp)

def pushfile(path):
  if path.count(':') > 0:
    host, path = path.split(':', 1)
  else:
    return
  tmp = 'delete_me-%x' % hash(path)
  if os.path.isfile(tmp):
    f = open(tmp+'_mtime', 'r')
    mt0 = f.read()
    f.close()
    if mt0 != str(os.path.getmtime(tmp)):
      out, err, duration, exitStatus = getProcessResults("scp %s %s:%s" % (tmp, host, path))
      if err: __abort("Unable to send \"%s:%s\"\n%s" % (host, path, err))
      if exitStatus != 0: __abort("Exit status %s indicates error sending \"%s:%s\"" % (exitStatus, host, path))
    os.remove(tmp)
    os.remove(tmp+'_mtime')

class FlashEntryPoint(EntryPointTemplate):
  def entryPoint1(self):
    log           = self.masterDict["log"]            # guaranteed to exist by flashTest.py
    flashTestOpts = self.masterDict["flashTestOpts"]  # guaranteed to exist by flashTest.py

    ##############################
    ##  CHECK FOR FLASH SOURCE  ##
    ##############################

    pathToFlash = flashTestOpts.get("-z","")
    if not pathToFlash:
      # check if we got a path to FLASH from the config file
      pathToFlash = self.masterDict.get("pathToFlash","")
      # make sure we have an absolute path
      if pathToFlash and not os.path.isabs(pathToFlash):
        pathToFlash = os.path.join(pathToFlashTest, pathToFlash)
    else:
      # make sure we have an absolute path
      if not os.path.isabs(pathToFlash):
        pathToFlash = os.path.join(os.getcwd(), pathToFlash)

    if not pathToFlash:
      log.err("You must provide a path to a copy of the FLASH source\n" +
              "either in a \"config\" file or with the \"-z\" option.")
      return False
    elif not os.path.isdir(pathToFlash):
      if "-u" in flashTestOpts:
        log.warn("\"%s\" does not exist or is not a directory." % pathToFlash)
      else:
        log.err("\"%s\" does not exist or is not a directory." % pathToFlash)

    self.masterDict["pathToFlash"] = pathToFlash

    ####################################
    ##  UPDATE FLASH SOURCE IF ASKED  ##
    ####################################

    pathToInvocationDir = self.masterDict["pathToInvocationDir"]  # guaranteed to exist by flashTest.py

    if "-u" in flashTestOpts:

      updateScript = self.masterDict.get("updateScript","").strip()
      if updateScript:

        # Capture the differences between tester's local working copy
        # and the HEAD revision unless those differences are too long.
        diffFlash = self.masterDict.get("diffFlash","").strip()
        if diffFlash:
          log.stp("Attempting to 'diff' this working copy and the HEAD repository revision\n" +
                  "with \"%s\"" % diffFlash)
          diffOut = ("output from \"%s\":\n" % diffFlash +
                     "**************" + "*"*len(diffFlash) + "\n")
          # We have to use popen for this instead of "getProcessResults" because
          # we're redirecting the output of the "diffFlash" command through a pipe,
          # and the pseudoterminals of "getProcessResults" don't deal well with those.
          p = subprocess.Popen([diffFlash + " | wc"],shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE,close_fds=True)
          p.wait()
          out = p.stdout.read().strip()
          size = int(out.split()[2])
          if size < 10000000:  # ~10 MB
            # the word count from "wc" was ok, so we can go back to using "getProcessResults"
            diffOut += (getProcessResults(diffFlash)[0] + "\n")
          else:
            diffOut += "Byte count from <diffFlash> is too long. Skipping this step.\n"
        else:
          diffOut = ""

        log.stp("Attempting to update FLASH source at \"%s\" with \"%s\"" % (pathToFlash, updateScript))

        cwd = os.getcwd()
        if (updateScript[0:3] != "rm "): os.chdir(pathToFlash)

        out, err, duration, exitStatus = getProcessResults(updateScript)
        if exitStatus != 0:
          open(os.path.join(pathToInvocationDir, "update_output"),
               "w",
          ).write(diffOut)
          if (updateScript[0:3] != "rm "): os.chdir(cwd)
          log.err("Exit status %s indicates error updating FLASH source..." % exitStatus)
          if (err != ""):
            log.info("stderr: " + err)
          if (out != "") and (out != err):
            log.info("stdout: " + out)
          return False
        else:
          log.info("FLASH source was successfully updated")
          open(os.path.join(pathToInvocationDir, "update_output"),
               "w",
          ).write(diffOut + out)

        os.chdir(cwd)
      else:
        log.err("\"-u\" passed to command line but no key \"updateScript\"\n" +
                "found in \"config\". Unable to update FLASH source")
        return False
    else:
      log.warn("FLASH source at \"%s\" was not updated" % pathToFlash)

    return True


  def entryPoint2(self):
    testPath = self.masterDict["testPath"]  # guaranteed to exist by flashTest.py
    firstElement = testPath.split("/",1)[0]

    # If this is a Default test, give it a special setupper component
    # (will be be automatically installed based on value in 'masterDict')
    if firstElement == "Default":
      self.masterDict["setupper"] = "DefaultSetupper"

    # Write some info into "linkAttributes" for use by FlashTestView
    if firstElement == "Comparison" or "Composite":
      pathToBuildDir = self.masterDict["pathToBuildDir"]  # guaranteed to exist by flashTest.py
      tester = self.masterDict.get("tester", "SfocuTester")
      text = "testerClass: %s" % tester
      _write_to_file(os.path.join(pathToBuildDir, "linkAttributes"),
                     "w",
                     text,
                     errors="backslashreplace")


  def entryPoint3(self):
    testPath = self.masterDict["testPath"]  # guaranteed to exist by flashTest.py
    firstElement = testPath.split("/",1)[0]

    # give this test an appropriate executer component (will be
    # be automatically installed based on value in 'masterDict')
    if firstElement == "Comparison":
      self.masterDict["executer"] = "ComparisonExecuter"
    elif firstElement == "Restart":
      self.masterDict["executer"] = "RestartExecuter"
    elif firstElement == "Composite":
      self.masterDict["executer"] = "CompositeExecuter"

    # give this test an appropriate tester component (will be
    # be automatically installed based on value in 'masterDict')
    if firstElement == "Comparison":
      tester = self.masterDict.get("tester")
      if tester == "GridDumpTester":
        self.masterDict["tester"] = "GridDumpTester"
      else:
        self.masterDict["tester"] = "SfocuTester"
    elif firstElement == "Restart":
      self.masterDict["tester"] = "RestartTester"
    elif firstElement == "UnitTest":
      self.masterDict["tester"] = "UnitTester"
    elif firstElement == "Composite":
      tester = self.masterDict.get("tester")
      if tester == "CompositeTesterExperimentalIO":
        self.masterDict["tester"] = "CompositeTesterExperimentalIO"
      else:
        self.masterDict["tester"] = "CompositeTester"


class FlashSetupper(SetupperTemplate):
  def setup(self):
    """
    run the FLASH setup script

                log: pointer to FlashTest logfile object
        pathToFlash: abs path to top-level FLASH directory
                     containing all code pertaining to FLASH setups
     pathToBuildDir: abs path to the output dir for this build
    pathToFlashTest: abs path to the top-level FlashTest directory
          setupName: name of this Flash setup (Sod, Sedov, etc.)
       setupOptions: options passed to Flash setup script
    """
    log             = self.masterDict["log"]              # guaranteed to exist by flashTest.py
    pathToFlash     = self.masterDict["pathToFlash"]      # guaranteed to exist by flashTest.py
    pathToBuildDir  = self.masterDict["pathToBuildDir"]   # guaranteed to exist by flashTest.py
    pathToFlashTest = self.masterDict["pathToFlashTest"]  # guaranteed to exist by flashTest.py

    setupName      = self.masterDict.get("setupName","")
    setupOptions   = self.masterDict.get("setupOptions","")
    #providing a more general way of choosing a nondefault makefile -PR
    flashMakefile  = self.masterDict.get("flashMakefile","")

    if len(setupName) == 0:
      log.err("No setup name provided.\n" +
              "Skipping this build.")
      return False

    #if this is defined, go with nondefault makefile.
    if len(flashMakefile) != 0:
      setupOptions += " -makefile="
      setupOptions += flashMakefile

    pathToDotSuccess = os.path.join(pathToFlash, "object", ".success")
    if os.path.isfile(pathToDotSuccess):
      os.remove(pathToDotSuccess)

    # setup script
    pathToFlashSetupScript = os.path.join(pathToFlash, self.masterDict.get("setupExec","setup"))
    script = "%s %s %s" % (pathToFlashSetupScript, setupName, setupOptions)

    # record setup invocation
    _write_to_file(os.path.join(pathToBuildDir, "setup_call"), "w",
                   script,
                   errors="backslashreplace")

    # log timestamp of command
    log.stp(script)

    # cd to FLASH source
    os.chdir(pathToFlash)

    # get stdout/stderr and duration of setup and write to file
    out, err, duration, exitStatus = getProcessResults(script)
    io.open(os.path.join(pathToBuildDir, "setup_output"), "w",
            errors="backslashreplace",
    ).write(out)
    if len(err) > 0:
        io.open(os.path.join(pathToBuildDir, "setup_error"), "w",
                errors="backslashreplace",
        ).write(err)

    # cd back to flashTest
    os.chdir(pathToFlashTest)

    # return the success or failure of the setup
    if os.path.isfile(pathToDotSuccess):
      log.stp("setup was successful")
      os.remove(pathToDotSuccess)
      return True
    else:
      log.stp("setup was not successful")
      return False


class FlashCompiler(CompilerTemplate):
  def compile(self):
    """
    compile FLASH

       pathToFlash: abs path up to the top-level FLASH directory
    pathToBuildDir: abs path up to the output dir for this build
       pathToGmake: abs path to the gmake utility
           exeName: name to be given to the FLASH executable
    """
    log             = self.masterDict["log"]              # guaranteed to exist by flashTest.py
    pathToFlash     = self.masterDict["pathToFlash"]      # guaranteed to exist by flashTest.py
    pathToBuildDir  = self.masterDict["pathToBuildDir"]   # guaranteed to exist by flashTest.py
    pathToFlashTest = self.masterDict["pathToFlashTest"]  # guaranteed to exist by flashTest.py

    pathToGmake    = self.masterDict.get("pathToGmake", "gmake")
    exeName        = self.masterDict.get("exeName", "flash-exe")

    pathToDotSuccess = os.path.join(pathToFlash, "object", ".success")
    if os.path.isfile(pathToDotSuccess):
      os.remove(pathToDotSuccess)

    # determine gmake invocation and record it in "gmake_call" file and in log
    script = "%s EXE=%s" % (pathToGmake, os.path.join(pathToBuildDir, exeName))
    _write_to_file(os.path.join(pathToBuildDir, "gmake_call"), "w",
                   script,
                   errors="backslashreplace")
    log.stp(script)

    # we'll try to compile multiple times if the compilation fails
    # because of a problem with the license manager
    numTries = 0
    mode = "w"

    while numTries < 3:

      # cd to FLASH source object directory for compilation
      os.chdir(os.path.join(pathToFlash, "object"))

      # get stdout/stderr and duration of compilation and write to file
      out, err, duration, exitStatus = getProcessResults(script)
      io.open(os.path.join(pathToBuildDir, "gmake_output"), mode,
              encoding="utf-8", errors="backslashreplace",
      ).write(out)
      if len(err) > 0:
          io.open(os.path.join(pathToBuildDir, "gmake_error"), mode,
                  errors="backslashreplace",
          ).write(err)

      # cd back to flashTest
      os.chdir(pathToFlashTest)

      # return the success or failure of the compilation

      if os.path.isfile(pathToDotSuccess):
        log.stp("compilation was successful")
        # record compilation time in "compilation_time" file and in log
        duration = secondsToHuman.convert(duration)
        open(os.path.join(pathToBuildDir, "compilation_time"),
             "w",
        ).write(duration)
        log.info("duration of compilation: %s" % duration)
        self.masterDict["pathToFlashExe"] = os.path.join(pathToBuildDir, exeName)  # for exeScript
        os.remove(pathToDotSuccess)
        return True
      elif out.find("www.macrovision.com") > 0:
        log.stp("compilation failed due to licensing problem. Trying again...")
        numTries += 1
      elif out.find("Unable to obtain license: license not found") > 0:
        log.stp("compilation failed due to license problem. Trying again...")
        numTries += 1
        mode = "a"
      else:
        break

    log.stp("compilation was not successful")
    return False

  def getDeletePatterns(self):
    return [self.masterDict.get("exeName", "flash-exe")]


class FlashExecuter(ExecuterTemplate):
  def execute(self, timeout=None):
    """
    run the FLASH executable piping output and other data into 'runDir'

    pathToRunDir: abs path to output dir for this unique executable/parfile combination
        numProcs: number of processors used for this run
         parfile: name of the parfile to be used in this run
    """
    log             = self.masterDict["log"]              # guaranteed to exist by flashTest.py
    pathToFlash     = self.masterDict["pathToFlash"]      # guaranteed to exist by flashTest.py
    pathToRunDir    = self.masterDict["pathToRunDir"]     # guaranteed to exist by flashTest.py
    pathToFlashTest = self.masterDict["pathToFlashTest"]  # guaranteed to exist by flashTest.py
    pathToExeScript = self.masterDict["pathToExeScript"]  # guaranteed to exist by flashTest.py

    # read the execution script from "exeScript"
    exeScriptFile = os.path.join(pathToFlashTest, pathToExeScript)
    if not os.path.isfile(exeScriptFile):
      log.err("File \"exeScript\" not found. Unable to run executable.\n" +
              "Skipping all runs.")
      return False

    lines = open(exeScriptFile).read().split("\n")
    lines = [line.strip() for line in lines
             if len(line.strip()) > 0 and not line.strip().startswith("#")]
    script = "\n".join(lines)
    self.masterDict["script"] = script
    script = self.masterDict["script"]  # do it this way so that any angle-bracket variables
                                        # in "exeScript" will be filled in by self.masterDict
    # determine 'pathToRunSummary'
    pathToRunSummary = os.path.join(pathToRunDir, "run_summary")

    # cd to output directory to run executable
    os.chdir(pathToRunDir)

    # obtain and record number of processors
    if not self.masterDict.has_key("numProcs"):
      self.masterDict["numProcs"] = 1
    open(pathToRunSummary,
         "a",
    ).write("numProcs: %s\n" % self.masterDict["numProcs"])

    # record mpirun invocation in "flash_call" file and in log
    _write_to_file(os.path.join(pathToRunDir, "flash_call"),
                   "w",
                   script,
                   errors="backslashreplace",
    )
    log.stp(script)

    # get stdout/stderr and duration of execution and write to file
    out, err, duration, exitStatus = getProcessResults(script, timeout)

    io.open(os.path.join(pathToRunDir, "flash_output"), "a",
            errors="ignore",
    ).write(out)
    if len(err) > 0:
        io.open(os.path.join(pathToRunDir, "flash_error"), "a",
                errors="backslashreplace",
        ).write(err)

    # record execution time in the run summary and logfile in human-readable form
    duration = secondsToHuman.convert(duration)
    open(pathToRunSummary, "a").write("wallClockTime: %s\n" % duration)
    log.info("duration of execution: %s" % duration)

    # search the parfile output directory for checkpoint files
    checkFiles = []
    items = os.listdir(pathToRunDir)
    for item in items:
      if re.match(".*?_chk_\d+$", item):
        checkFiles.append(item)

    # record number and names of checkpoint files in the run summary
    open(pathToRunSummary, "a").write("numCheckfiles: %s\n" % len(checkFiles))
    for checkFile in checkFiles:
        open(pathToRunSummary, "a").write("checkFile: %s\n" % checkFile)

    # An exit status of 0 means a normal termination without errors.
    if exitStatus == 0:
      log.stp("Process exit-status reports execution successful")
      runSucceeded = True
    else:
      log.stp("Process exit-status reports execution failed")
      pathToFlashExe = self.masterDict["pathToFlashExe"]  # set by FlashCompiler, above
      if len(pathToFlashExe) > 0 and pathToFlashExe[0] == '/':
        time.sleep(1)
        killallCmd = 'killall -v -HUP %s' % pathToFlashExe
#        log.stp('killallCmd is "%s"' % killallCmd)
        p = subprocess.Popen([killallCmd],shell=True,stderr=subprocess.PIPE,close_fds=True)
        killStatus = p.wait()
        if killStatus==0:
          out = p.stderr.read().strip()
          if len(out) > 0:
            log.stp(killallCmd)
            log.info(out)
#        else:
#          log.stp("Exit status %d from %s" % (killStatus, killallCmd))
      else:
        log.stp('pathToFlashExe is "%s"' % pathToFlashExe)
      runSucceeded = False

    # cd back to flashTest
    os.chdir(pathToFlashTest)

    return runSucceeded

  def getDeletePatterns(self):
    deletePatterns = self.masterDict.get("deletePatterns", [])
    if len(deletePatterns) > 0:
      # split 'deletePatterns' into a list on whitespace
      deletePatterns = re.split("\s+",deletePatterns.strip())

    deletePatterns.extend([".*_chk_\d+$", ".*_plt_cnt_\d+$"])
    return deletePatterns


class ComparisonExecuter(FlashExecuter):
  def adjustFilesToDelete(self, filesToDelete):
    """
    Determine the highest-numbered checkpoint file and create an
    entry "chkMax" in masterDict whose value is the name of this
    file. We'll use this value later to do our sfocu comparison.

    Then remove this file's name from 'filesToDelete', which will
    later be used to determine which files will be deleted before
    creation of the slim copy of the invocation's output.
    """
    pathToRunDir = self.masterDict["pathToRunDir"]  # guaranteed to exist by flashTest.py
    chkFiles = []

    # Search 'runDir' for checkpoint files. This method will also
    # be called for GridDumpComparison problems that do not generate
    # checkpoint files, but nothing will happen in that case.
    items = os.listdir(pathToRunDir)
    for item in items:
      if re.match(".*?_chk_\d+$", item):
        chkFiles.append(item)

    # sorting and reversing will put the highest-numbered
    # checkpoint file at index 0
    chkFiles.sort()
    chkFiles.reverse()

    if len(chkFiles) > 0:
      chkMax = chkFiles[0]
      self.masterDict["chkMax"] = chkMax
      for fileToDelete in filesToDelete[:]:
        if fileToDelete == chkMax:
          filesToDelete.remove(fileToDelete)

class ComparisonTester(TesterTemplate):

  def compare(self, pathToFileA, pathToFileB, cmd):
    log                = self.masterDict["log"]                # guaranteed to exist by flashTest.py
    arch               = self.masterDict["arch"]               # guaranteed to exist by flashTest.py
    outfile            = self.masterDict["outfile"]            # guaranteed to exist by flashTest.py
    pathToRunDir       = self.masterDict["pathToRunDir"]       # guaranteed to exist by flashTest.py
    pathToLocalArchive = self.masterDict["pathToLocalArchive"] # guaranteed to exist by flashTest.py

    pathToFileA = os.path.normpath(pathToFileA)
    pathToFileB = os.path.normpath(pathToFileB)

    if not os.path.isabs(pathToFileA):
      pathToFileA = os.path.join(pathToRunDir, pathToFileA)
    if not os.path.isabs(pathToFileB):
      pathToFileB = os.path.join(pathToRunDir, pathToFileB)

    if pathToFileA.startswith(pathToLocalArchive):
      try:
        arch.confirmInLocalArchive(pathToFileA)
      except Exception as e:
        log.err("%s\n" % e +
                "Aborting this test.")
        outfile.write(str(e))
        return False
    elif not os.path.isfile(pathToFileA):
      if not os.path.exists(pathToFileA):
        log.stp("\"%s\" does not exist." % pathToFileA)
        outfile.write("\"%s\" does not exist.\n" % pathToFileA)
      else:
        log.stp("\"%s\" is not a regular file." % pathToFileA)
        outfile.write("\"%s\" is not a regular file.\n" % pathToFileA)
      return False

    if pathToFileB.startswith(pathToLocalArchive):
      try:
        arch.confirmInLocalArchive(pathToFileB)
      except Exception as e:
        log.err("%s\n" % e +
                "Aborting this test.")
        outfile.write(str(e))
        return False
    elif not os.path.isfile(pathToFileB):
      if not os.path.exists(pathToFileB):
        log.stp("\"%s\" does not exist." % pathToFileB)
        outfile.write("\"%s\" does not exist.\n" % pathToFileB)
      else:
        log.stp("\"%s\" is not a regular file." % pathToFileB)
        outfile.write("\"%s\" is not a regular file.\n" % pathToFileB)
      return False

    log.stp("FileA: \"%s\"\n" % pathToFileA +
            "FileB: \"%s\""   % pathToFileB)
    outfile.write("FileA: \"%s\"\n" % pathToFileA +
                  "FileB: \"%s\"\n\n" % pathToFileB)

    outfile.write("script: %s\n" % cmd)
    return getProcessResults(cmd)


  def compareToYesterday(self, pathToFile, pathToCompareExecutable):
    yesterDate = time.strftime("%Y-%m-%d", time.localtime(time.time()-24*60*60))

    pat1 = re.compile("\/\d\d\d\d-\d\d-\d\d.*?\/")
    pathToYesterFile = pat1.sub("/%s/" % yesterDate, pathToFile)

    cmd = "%s %s %s" % (pathToCompareExecutable, pathToFile, pathToYesterFile)
    return self.compare(pathToFile, pathToYesterFile, cmd)


  def yesterFileIsSameFile(self, pathToFile):
    # Try to detect situation where the putative file from yesterday is really
    # this run's file.  This happens when the current invocation was started
    # on the previous day and passed midnight.  The simple algorithm based on
    # localtime() will then fail to locate the file from the previous run correctly. - KW
    yesterDate = time.strftime("%Y-%m-%d", time.localtime(time.time()-24*60*60))

    pat1 = re.compile("\/\d\d\d\d-\d\d-\d\d.*?\/")
    pathToYesterFile = pat1.sub("/%s/" % yesterDate, pathToFile)

    return os.path.samefile(pathToFile, pathToYesterFile)


class SfocuTester(ComparisonTester):

  def test(self):
    log          = self.masterDict["log"]           # guaranteed to exist by flashTest.py
    pathToFlash  = self.masterDict["pathToFlash"]   # guaranteed to exist by flashTest.py
    pathToRunDir = self.masterDict["pathToRunDir"]  # guaranteed to exist by flashTest.py
    outfile      = self.masterDict["outfile"]       # guaranteed to exist by flashTest.py

    if not self.masterDict.has_key("chkMax"):
      log.stp("No checkpoint files were produced, so no comparisons can be made.")
      outfile.write("No checkpoint files were produced, so no comparisons can be made.\n")
      return False

    # else
    pathToChkMax = os.path.join(pathToRunDir, self.masterDict["chkMax"])
    pathToSfocu = self.masterDict.get("pathToSfocu", os.path.join(pathToFlash, "tools", "sfocu", "sfocu"))
    sfocuScript = self.masterDict.get("sfocuScript", pathToSfocu)

    # before comparing to the benchmark, compare to yesterday's result
    # this portion assumes that copies of the highest-numbered checkfile
    # are being retained locally in FlashTest's "output" directory
    log.stp("Part 1: Compare this invocation's result to yesterday's")
    outfile.write("Part 1: Compare this invocation's result to yesterday's\n")

    retval = self.compareToYesterday(pathToChkMax, sfocuScript)

    if retval:
      # unpack the tuple
      out, err, duration, exitStatus = retval

      # An exit status of 0 means a normal termination without errors.
      if exitStatus == 0:
        log.stp("Process exit-status reports sfocu ran successfully")
        outfile.write("<b>sfocu output:</b>\n"
                      + out.strip() + "\n\n")

        # Even if sfocu ran fine, the test might still have failed
        # if the two checkpoint files were not equivalent
        if out.strip().endswith("SUCCESS"):
          log.stp("comparison of benchmark files yielded: SUCCESS")
          # Set sameAsPrevious, but only after some additonal checking:
          # - The checkpoint we are comparing must be numbered higher than 0.
          # - Comparison against "yesterday's" checkpoint must have really
          #   been a comparison with a DIFFERENT file.
          if (self.masterDict["chkMax"][-5:-1] != "_0000") and \
             not self.yesterFileIsSameFile(pathToChkMax):
            self.masterDict["sameAsPrevious"] = True
        else:
          log.stp("comparison of benchmark files yielded: FAILURE")
          # The results of this test differed from the results of the
          # same test done during the previous invocation. We set the
          # key "changedFromPrevious" in masterDict (the value doesn't
          # matter) which is recognized by flashTest.py as a signal to
          # add a "!" to the ends of the "errors" files at the run,
          # build, and invocation levels.
          self.masterDict["changedFromPrevious"] = True
      else:
        log.stp("Process exit-status reports sfocu encountered an error")

        # record whatever we got anyway
        outfile.write("Process exit-status reports sfocu encountered an error\n" +
                      "<b>sfocu output:</b>\n" +
                      out.strip() + "\n\n")

    log.stp("Part 2: Compare this invocation's result to approved benchmark.")
    outfile.write("Part 2: Compare this invocation's result to approved benchmark.\n")

    if not self.masterDict.has_key("shortPathToBenchmark"):
      log.err("A key \"shortPathToBenchmark\", whose value is a relative path from\n" +
              "the local archive to a benchmark file against which the results of\n" +
              "this run can be compared, should be provided in your \"test.info\" file.")
      return False

    # else
    shortPathToBenchmark = self.masterDict["shortPathToBenchmark"]
    pathToLocalArchive = self.masterDict["pathToLocalArchive"]  # guaranteed to exist by flashTest.py
    pathToBenchmark = os.path.join(pathToLocalArchive, shortPathToBenchmark)

    cmdAndTols = [sfocuScript]
    if self.masterDict.has_key("errTol"):
      cmdAndTols.append("-e %s" % self.masterDict["errTol"])
    if self.masterDict.has_key("partErrTol"):
      cmdAndTols.append("-p %s" % self.masterDict["partErrTol"])

    cmd = "%s %s %s" % (" ".join(cmdAndTols), pathToChkMax, pathToBenchmark)
    retval = self.compare(pathToChkMax, pathToBenchmark, cmd)

    if not retval:
      return False

    # else unpack the tuple
    out, err, duration, exitStatus = retval

    # An exit status of 0 means a normal termination without errors.
    if exitStatus == 0:
      log.stp("Process exit-status reports sfocu ran successfully.")
      outfile.write("<b>sfocu output:</b>\n"
                    + out.strip() + "\n\n")

      # Even if sfocu ran fine, the test might still have failed
      # if the two checkpoint files were not equivalent
      if out.strip().endswith("SUCCESS"):
        log.stp("comparison of benchmark files yielded: SUCCESS")
        return True
      else:
        log.stp("comparison of benchmark files yielded: FAILURE")
        return False
    else:
      log.stp("Process exit-status reports sfocu encountered an error")

      # record whatever we got anyway
      outfile.write("Process exit-status reports sfocu encountered an error\n" +
                    "<b>sfocu output:</b>\n" +
                    out.strip() + "\n\n")
      # sfocu had an error, so we return false:
      return False

    return True


class GridDumpTester(ComparisonTester):
  def findDumpFiles(self):
    pathToRunDir = self.masterDict["pathToRunDir"]  # guaranteed to exist by flashTest.py

    dumpFiles = []
    items = os.listdir(pathToRunDir)
    for item in items:
      if re.match("^FL\d+$", item):
        dumpFiles.append(item)

    return dumpFiles

  def test(self):
    log          = self.masterDict["log"]           # guaranteed to exist by flashTest.py
    pathToFlash  = self.masterDict["pathToFlash"]   # guaranteed to exist by flashTest.py
    pathToRunDir = self.masterDict["pathToRunDir"]  # guaranteed to exist by flashTest.py
    outfile      = self.masterDict["outfile"]       # guaranteed to exist by flashTest.py

    pathToGridDumpCompare = self.masterDict.get("pathToGridDumpCompare",
                                                os.path.join(pathToFlash, "tools", "GridDumpCompare.py"))
    if not os.path.isfile(pathToGridDumpCompare):
      log.err("\"%s\" does not exist or is not a regular file.\n" % pathToGridDumpCompare +
              "Result of test: FAILURE")
      outfile.write("\"%s\" does not exist or is not a regular file.\n" % pathToGridDumpCompare +
                    "Result of test: FAILURE\n")
      return False

    # else
    dumpFiles = self.findDumpFiles()

    if len(dumpFiles) == 0:
      log.err("No GridDump files were found, so no comparison can be made." +
              "Result of test: FAILURE")
      outfile.write("No GridDump files were found, so no comparison can be made.\n" +
                    "Result of test: FAILURE\n")
      return False

    # else
    log.stp("Part 1: Compare this invocation's results to yesterday's")
    outfile.write("Part 1: Compare this invocation's results to yesterday's\n")

    for dumpFile in dumpFiles:
      pathToDumpFile = os.path.join(pathToRunDir, dumpFile)

      retval = self.compareToYesterday(pathToDumpFile, pathToGridDumpCompare)

      if retval:
        # unpack the tuple
        out, err, duration, exitStatus = retval

        if exitStatus == 0:
          log.stp("comparison of dump-files yielded: SUCCESS")
        elif exitStatus == 1:
          log.stp("comparison of dump-files yielded: FAILURE")
          # The results of this test differed from the results of the
          # same test done during the previous invocation. We set the
          # key "changedFromPrevious" in masterDict (the value doesn't
          # matter) which is recognized by flashTest.py as a signal to
          # add a "!" to the ends of the "errors" files at the run,
          # build, and invocation levels.
          self.masterDict["changedFromPrevious"] = True
        else:
          log.stp("Process exit-status reports GridDumpCompare.py encountered an error")

        # record whatever we got in all cases
        outfile.write("<b>GridDumpCompare.py output:</b>\n" +
                      out.strip() + "\n\n")

    log.stp("Part 2: Compare this invocation's result to approved benchmark.")
    outfile.write("Part 2: Compare this invocation's result to approved benchmark.\n")

    if not self.masterDict.has_key("shortPathToBenchmarkDir"):
      log.err("A key \"shortPathToBenchmarkDir\", whose value is a relative path from the\n" +
              "local archive to the directory containing the files against which the results\n" +
              "of this run can be compared, should be provided in your \"test.info\" file.")
      return False

    # else
    # For GridDump comparisons we bring over a whole directory from the archive
    # which contains all dumps made for a single run (each dump encapsulates the
    # values of a single variable at different points on the grid).
    # This is different from sfocu comparisons, where we only bring over a single
    # file, the highest-numbered checkpoint file.
    shortPathToBenchmarkDir = self.masterDict["shortPathToBenchmarkDir"]

    arch = self.masterDict["arch"]  # guaranteed to exist by flashTest.py
    try:
      arch.confirmInLocalArchive(shortPathToBenchmarkDir)
    except Exception as e:
      log.err("%s\n" % e +
              "Aborting this test.")
      outfile.write(str(e))
      return False

    # else all files have been successfully brought over from
    # the remote archive into the local archive
    pathToLocalArchive = self.masterDict["pathToLocalArchive"]  # guaranteed to exist by flashTest.py

    allPassed = True

    for dumpFile in dumpFiles:
      pathToDumpFile = os.path.join(pathToRunDir, dumpFile)
      pathToBenchmark = os.path.join(pathToLocalArchive, shortPathToBenchmarkDir, dumpFile)

      cmd = "%s %s %s" % (pathToGridDumpCompare, pathToDumpFile, pathToBenchmark)
      retval = self.compare(pathToDumpFile, pathToBenchmark, cmd)

      if not retval:
        allPassed = False
        continue

      # else unpack the tuple
      out, err, duration, exitStatus = retval

      if exitStatus == 0:
        log.stp("comparison of dump-files yielded: SUCCESS")
      elif exitStatus == 1:
        log.stp("comparison of dump-files yielded: FAILURE")
        allPassed = False
      else:
        log.stp("Process exit-status reports GridDumpCompare.py encountered an error")
        allPassed = False

      # record whatever we got in all cases
      outfile.write("<b>GridDumpCompare.py output:</b>\n" +
                    out.strip() + "\n\n")

    return allPassed


class UnitTester(TesterTemplate):
  """
  Implements a test method for all Flash unit-tests that follow the accepted
  unit-test standard. This standard prescribes that a unit-test will produce
  files named:

    unitTest_0, unitTest_1, ..., unitTest_n

  where 'n' is the number of processors on which the test ran. If the unit-
  test was completely successful, each of these files will contain the text:

    "all results conformed with expected values."

  Otherwise they will contain information describing why the test failed.

  This test method reads files whose names match those described above and
  determines success or failure based on the presence or absence therein of
  the aforementioned success string.
  """
  def test(self):
    log          = self.masterDict["log"]           # guaranteed to exist by flashTest.py
    pathToFlash  = self.masterDict["pathToFlash"]   # guaranteed to exist by flashTest.py
    pathToRunDir = self.masterDict["pathToRunDir"]  # guaranteed to exist by flashTest.py
    outfile      = self.masterDict["outfile"]       # guaranteed to exist by flashTest.py

    UGPat = re.compile("^unitTest_\d+$")

    files = os.listdir(pathToRunDir)
    files = [f for f in files if UGPat.search(f)]  # only want files
                                                   # matching 'UGPat'

    if len(files) > 0:
      success = True
      successStr = "all results conformed with expected values."
      # put in alphabetical order
      files.sort()
      for f in files:
        text =  "<i>reading file %s:</i>" % f
        text += "<div style='margin-left:10px;'>"
        fileText = open(os.path.join(pathToRunDir, f), "r").read().lower()
        if fileText.count(successStr) == 0:
          success = False
        text += fileText
        text += "</div>"
        outfile.write(text)
    else:
      success = False

    if success:
      log.stp("result of test: SUCCESS")
    else:
      log.stp("result of test: FAILURE")

    return success


class DefaultSetupper(FlashSetupper):
  def setup(self):
    """
    The Default test by definition sets up the problem called "Default"
    This test doesn't use a "test.info" file, so we override the setup
    method long enough to set some key values in 'masterDict', then
    revert back to the setup method of UniversalTemplate.
    """
    self.masterDict["setupName"] = "Default"
    self.masterDict["setupOptions"] = "-site=%s -auto" % self.masterDict["flashSite"]

    return FlashSetupper.setup(self)

class SiteSetupper(FlashSetupper):
  def setup(self):
    """
    This setupper differs from the normal behavior only by prepending
    " -site=<flashSite>" to the setupOptions and thus to the command line for the
    setup call.
    """
    if self.masterDict.get("flashSite",""):
      setupOptions   = self.masterDict.get("setupOptions","")
      self.masterDict["setupOptions"] = "-site=%s %s" % (self.masterDict["flashSite"], setupOptions)

    return FlashSetupper.setup(self)

class IntelSetupper(SiteSetupper):
  def setup(self):
    """
    This setupper differs from the SiteSettuper only by prepending
    " -makefile=Intel" to the setupOptions and thus to the command line for the
    setup call.
    """
    setupOptions   = self.masterDict.get("setupOptions","")
    self.masterDict["setupOptions"] = "-makefile=Intel %s" % setupOptions

    return SiteSetupper.setup(self)

class DebugSetupper(FlashSetupper):
  def setup(self):
    """
    This setupper differs from the normal behavior only by appending
    " -debug" to the setupOptions and thus to the command line for the
    setup call.
    """
    setupOptions   = self.masterDict.get("setupOptions","")
    self.masterDict["setupOptions"] = "%s -debug" % setupOptions
    self.masterDict["usingDebugSetupper"] = True

    return FlashSetupper.setup(self)

class LibmodeSetupper(FlashSetupper):
  def setup(self):
    """
    This setupper differs from the normal behavior only by appending
    " ParameshLibraryMode=True" to the setupOptions and thus to the command line
    for the setup call.
    """
    setupOptions   = self.masterDict.get("setupOptions","")
    self.masterDict["setupOptions"] = "%s ParameshLibraryMode=True" % setupOptions

    return FlashSetupper.setup(self)

class NoClobberSetupper(SiteSetupper):
  def setup(self):
    """
    This setupper differs from the SiteSetupper only by appending
    " -noclobber" to the setupOptions and thus to the command line for the
    setup call.
    """
    setupOptions   = self.masterDict.get("setupOptions","")
    self.masterDict["setupOptions"] = "%s -noclobber" % setupOptions

    return SiteSetupper.setup(self)

class StaticHyArraysSetupper(NoClobberSetupper):
  def setup(self):
    """
    This setupper differs from the SiteSetupper only by appending
    " -noclobber" to the setupOptions and thus to the command line for the
    setup call.
    """
    setupOptions   = self.masterDict.get("setupOptions","")
    self.masterDict["setupOptions"] = "%s StaticHyArrays=True" % setupOptions

    return NoClobberSetupper.setup(self)


class RestartExecuter(ComparisonExecuter):
  """
  This class subclasses ComparisonExecuter instead of FlashExecuter
  so that the former's "adjustFilesToDelete()" method, which sets a
  value for "checkMax" will be called.
  """
  def execute(self):
    """
    The Restart test first runs a simulation in the normal way, generating
    a few checkpoint files. It then takes a checkpoint file from the middle
    of the run (as specified in the "test.info" data) and runs the problem
    again, this time restarting from the checkpoint. If the end-checkpoint
    file from the restart-run matches the end-checkpoint file from the first
    run, the test is considered a success
    """
    log          = self.masterDict["log"]           # guaranteed to exist by flashTest.py
    pathToRunDir = self.masterDict["pathToRunDir"]  # guaranteed to exist by flashTest.py

    pathToPart1Parfile  = self.masterDict["part1Parfile"]   # supposed to exist in "test.info"
    part1NumProcs       = self.masterDict["part1NumProcs"]  # supposed to exist in "test.info"
    benchmark1          = self.masterDict["benchmark1"]     # supposed to exist in "test.info"
    benchmark2          = self.masterDict["benchmark2"]     # supposed to exist in "test.info"
    pathToExeScript = self.masterDict["pathToExeScript"]  # guaranteed to exist by flashTest.py

    # First run the simulation through to generate
    # benchmarks 1 and 2 in the normal way
    pathToFlashTest = self.masterDict["pathToFlashTest"]  # guaranteed to exist by flashTest.py

    # read the execution script from "exeScript"
    exeScriptFile = os.path.join(pathToFlashTest, pathToExeScript)
    if not os.path.isfile(exeScriptFile):
      log.err("File \"exeScript\" not found. Unable to run executable.\n" +
              "Skipping all runs.")
      return False

    # else

    # hackily substitute into masterDict's "parfile" keyword the
    # name of the parfile we need to run this whole simulation
    # *without* a restart so it will in turn be substituted into
    # masterDict["script"] below.
    part1Parfile                = os.path.basename(pathToPart1Parfile)
    part2Parfile                = self.masterDict["parfile"]
    part2NumProcs               = self.masterDict["numProcs"]
    self.masterDict["parfile"]  = part1Parfile
    self.masterDict["numProcs"] = part1NumProcs

    lines = open(exeScriptFile).read().split("\n")
    lines = [line.strip() for line in lines
             if len(line.strip()) > 0 and not line.strip().startswith("#")]
    script = "\n".join(lines)
    self.masterDict["script"] = script
    script = self.masterDict["script"]  # do it this way so that any angle-bracket variables
                                        # in "exeScript" will be filled in by self.masterDict

    log.stp("Generating seed-benchmark \"%s\"\n" % benchmark1 +
            "and comparison-benchmark \"%s\"\n" % benchmark2 +
            "with parfile \"%s\" and numProcs=%s" % (part1Parfile, part1NumProcs))

    # write part1 parfile into 'runDir'
    parfileText = open(pathToPart1Parfile).read()
    open(os.path.join(pathToRunDir, part1Parfile), "w").write(parfileText)

    # cd to output directory to run executable
    os.chdir(pathToRunDir)

    # get stdout/stderr and duration of execution
    out, err, duration, exitStatus = getProcessResults(script)

    flashOutput = "Results of generation of seed and comparison checkpoint files:\n"
    flashOutput += (out + "\n")
    flashOutput += (("*" * 80) + "\n")
    _write_to_file(os.path.join(pathToRunDir, "flash_output"), "w",
                   flashOutput,
                   errors="ignore",
    )

    if len(err) > 0:
      flashError = "Error in generation of seed and comparison checkpoint files:\n"
      flashError += (err + "\n")
      flashError += (("*" * 80) + "\n")
      _write_to_file(os.path.join(pathToRunDir, "flash_error"), "w",
                     flashError,
                     errors="backslashreplace",
      )

    # An exit status of 0 means a normal termination without errors.
    if exitStatus == 0:
      log.stp("Process exit-status reports execution successful")
    else:
      log.stp("Process exit-status reports execution failed.")
      pathToFlashExe = self.masterDict["pathToFlashExe"]  # set by FlashCompiler, above
      if len(pathToFlashExe) > 0 and pathToFlashExe[0] == '/':
        time.sleep(1)
        killallCmd = 'killall -v -HUP %s' % pathToFlashExe
#        log.stp('killallCmd is "%s"' % killallCmd)
        p = subprocess.Popen([killallCmd],shell=True,stderr=subprocess.PIPE,close_fds=True)
        killStatus = p.wait()
        if killStatus==0:
          out = p.stderr.read().strip()
          if len(out) > 0:
            log.stp(killallCmd)
            log.info(out)
#        else:
#          log.stp("Exit status %d from %s" % (killStatus, killallCmd))
      else:
        log.stp('pathToFlashExe is "%s"' % pathToFlashExe)
      return False

    # search the parfile output directory for the
    # seed-checkpoint and comparison-checkpoint files
    items = os.listdir(pathToRunDir)
    for item in items:
      if item == benchmark1:
        # this will be the seed-checkpoint file
        break  # skip "else" clause below
    else:
      log.stp("Expected seed-checkpoint file \"%s\" was not generated.\n" % benchmark1 +
              "Skipping all runs.")
      return False

    # search the parfile output directory for the comparison checkpoint file
    for item in items:
      if item == benchmark2:
        # This will be the comparison-checkpoint file.
        # Rename it so it won't get overwritten in part2.
        os.rename(item, item + "_orig")
        break  # skip "else" clause below
    else:
      log.stp("Expected comparison checkpoint file \"%s\" was not generated.\n" % benchmark2 +
              "Skipping all runs.")
      return False

    # cd back to flashTest
    os.chdir(pathToFlashTest)

    # We've now guaranteed that the "seed" checkpoint file is in
    # place in 'runDir', so reset the parfile and numProcs values
    # and call the parent class's executer
    self.masterDict["parfile"]  = part2Parfile
    self.masterDict["numProcs"] = part2NumProcs
    return ComparisonExecuter.execute(self)


class RestartTester(ComparisonTester):
  def test(self):
    log                = self.masterDict["log"]           # guaranteed to exist by flashTest.py
    pathToFlash        = self.masterDict["pathToFlash"]   # guaranteed to exist by flashTest.py
    pathToRunDir       = self.masterDict["pathToRunDir"]  # guaranteed to exist by flashTest.py
    outfile            = self.masterDict["outfile"]       # guaranteed to exist by flashTest.py

    benchmark1 = self.masterDict["benchmark1"]  # supposed to exist in "test.info"
    benchmark2 = self.masterDict["benchmark2"]  # supposed to exist in "test.info"

    pathToSfocu = self.masterDict.get("pathToSfocu", os.path.join(pathToFlash, "tools", "sfocu", "sfocu"))
    sfocuScript = self.masterDict.get("sfocuScript", pathToSfocu)

    chkMax = self.masterDict["chkMax"]  # should always have a value because even if execution produced
                                        # no checkpoint files, the seed-checkpoint copied into 'runDir'
                                        # should have been tagged as "chkMax" in ComparisonExecuter's
                                        # "adjustFilesToDelete" method.

    if chkMax == benchmark1:
      log.stp("No additional checkpoint files were produced after the restart.\n" +
              "No comparison can be made.")
      return False

    # else
    pathToChkMax = os.path.join(pathToRunDir, chkMax)
    pathToBenchmark2 = os.path.join(pathToRunDir, (benchmark2 + "_orig"))

    cmdAndOpts = [sfocuScript]
    if "errTol" in self.masterDict:
      cmdAndOpts.append("-e %s" % self.masterDict["errTol"])
    if "partErrTol" in self.masterDict:
      cmdAndOpts.append("-p %s" % self.masterDict["partErrTol"])
    if "sfocuFlags" in self.masterDict:
      compFlags = ""
      compFlags = self.masterDict["sfocuFlags"]
      cmdAndOpts.append(compFlags)

    cmd = "%s %s %s" % (" ".join(cmdAndOpts), pathToChkMax, pathToBenchmark2)
    retval = self.compare(pathToChkMax, pathToBenchmark2, cmd)

    if not retval:
      log.stp("Error processing command \"%s\"\n" % cmd)
      return False

    # else unpack the tuple
    out, err, duration, exitStatus = retval

    # An exit status of 0 means a normal termination without errors.
    if exitStatus == 0:
      log.stp("Process exit-status reports sfocu ran successfully.")
      outfile.write("<b>sfocu output:</b>\n"
                    + out.strip() + "\n\n")

      # Even if sfocu ran fine, the test might still have failed
      # if the two checkpoint files were not equivalent
      if out.strip().endswith("SUCCESS"):
        log.stp("comparison of benchmark files yielded: SUCCESS")
        return True
      else:
        log.stp("comparison of benchmark files yielded: FAILURE")
        return False
    else:
      log.stp("Process exit-status reports sfocu encountered an error")

      # record whatever we got anyway
      outfile.write("Process exit-status reports sfocu encountered an error\n" +
                    "<b>sfocu output:</b>\n" +
                    out.strip() + "\n\n")
      # sfocu had an error, so we return false:
      return False

    return True


class CompositeExecuter(FlashExecuter):
  def __init__ (self, testObject):
    """
    Override the template's method to allow us to correctly handle restarts.
    """
    self.owner = testObject
    self.masterDict = testObject.masterDict

    self.restartParfiles = None
    self.parfilesLeft = True

  def adjustFilesToDelete(self, filesToDelete):
    """
    Notes the comparison benchmark.  Only removes from to delete list if sfocu
    comparison has not changed.  This value is also used for the sfocu compariosn.
    Make sure called afer comparison.
    """
    #comparisonFile = os.path.basename(self.masterDict["comparisonBenchmark"])
    #restartFile = os.path.basename(self.masterDict["restartBenchmark"])
    checkpointBasename = self.masterDict["checkpointBasename"]
    comparisonFile = checkpointBasename + self.masterDict["comparisonNumber"]
    restartFile = checkpointBasename + self.masterDict["restartNumber"]

    #Populate the chkMax so we can determine if the restart test actually ran.
    pathToRunDir = self.masterDict["pathToRunDir"] #flashTest.py has this
    chkFiles = []

    files = os.listdir(pathToRunDir)
    for file in files:
      if re.match(".*?_chk_\d+$", file):
        chkFiles.append(file)

    chkFiles.sort()
    chkFiles.reverse()
    #this is the furthest file we've gotten on this run
    if len(chkFiles) > 0:
      chkMax = chkFiles[0]
      self.masterDict["chkMax"] = chkMax

#    for chkFile in chkFiles:
#       filesToDelete.append(chkFile)

    #get plot and part files?

    #DEV: have to make sure comparitor is smart enough to actually figure this out.
    #Assume that we are keeping the file, can add back in later?
    #DEV: Just Keep Them for now.
    pahToRunDir = self.masterDict["pathToRunDir"]

    for fileToDelete in filesToDelete[:]:
      if fileToDelete == restartFile:
        filesToDelete.remove(fileToDelete)
      elif fileToDelete == comparisonFile:
        filesToDelete.remove(fileToDelete)

    return

  def execute(self):
    """
    Overridden to allow both Comparison and restart behaviors
    """
    log             = self.masterDict["log"]              # guaranteed to exist by flashTest.py
    pathToFlash     = self.masterDict["pathToFlash"]      # guaranteed to exist by flashTest.py
    pathToRunDir    = self.masterDict["pathToRunDir"]     # guaranteed to exist by flashTest.py
    pathToFlashTest = self.masterDict["pathToFlashTest"]  # guaranteed to exist by flashTest.py
    pathToExeScript = self.masterDict["pathToExeScript"]  # guaranteed to exist by flashTest.py

    #DEV:Make sure these are safe.
    comparisonParfile = self.masterDict["parfile"]    #new for this mode

    #Since we are restarting we have to run twice, with different parfiles.
    #This is called once for each parfile entry in a multi-parfile run, so this
    #will only iterate twice per exectuion.

    #populate the restart parfiles, if we don't have them. Head will be next to use.
    if self.restartParfiles is None:
      restartParfileNames = self.masterDict["restartParfiles"] #new for this mode: parfile for restarts
      self.restartParfiles = restartParfileNames.split(None)
      #sanity check.  Specification is that the number of entries in parfiles and restartParfiles is the same.
      if len(self.restartParfiles) != len(self.masterDict["parfiles"]):
        log.err("Mismatch in number of entries in parfiles and restartParfiles entries for test.")
        return False

    else:
      if type(self.restartParfiles) == type('abc'):
        self.restartParfiles= self.restartParfiles.split(None)

    restartParfile = self.restartParfiles[0]
    del self.restartParfiles[0]
    if len(self.restartParfiles) == 0:
      self.restartParfiles = None
  ######################################################
    iteration = 0
    #loop, requires three iterations:
    while(iteration < 3):
      #make sure that the parfile has correct path.
      if iteration == 0:
        if "restartBenchmark" in self.masterDict or restartParfile == "none":
          iteration += 1
          continue # skip iteration
        else: # must verify transparency restart for auto approved restart benchmark
          log.stp("restartBenchmark not present, performing long-run to verify transparent restart.\n")
          def readpar(path):
            f = open(path, "r")
            parms = dict(
                [[m.group(1).lower(), m.group(2)]
                 for m in
                     [re.match("\s*(\w+)\s*=\s*(\S|\S.*\S)\s*$", line)
                      for line in f.readlines()]
                 if m
                 ]
            )
            f.close()
            return parms
          def writepar(path, parms):
            f = open(path, "w")
            f.writelines(["%s = %s\n" % (k,v) for k,v in parms.items() ])
            f.close()
          pathToCompParfile = os.path.join(pathToRunDir, comparisonParfile)
          pathToRestParfile = os.path.join(pathToRunDir, restartParfile)
          parComp = readpar(pathToCompParfile)
          parRest = readpar(pathToRestParfile)
          if "nend" in parRest: parComp["nend"] = parRest["nend"]
          if "tmax" in parRest: parComp["tmax"] = parRest["tmax"]
          if "dr_shortenlaststepbeforetmax" in parRest: parComp["dr_shortenlaststepbeforetmax"] = parRest["dr_shortenlaststepbeforetmax"]
          pathToParfile = os.path.join(pathToRunDir, "restart_longrun_hyn7rt0v20.par")
          writepar(pathToParfile, parComp)
      elif iteration == 1:
        pathToParfile = comparisonParfile
      #we have to swap parfiles.  If none, note this fact.
      #if the none placeholder is used, stamp a message, not an
      #error.  This is intentional.
      elif iteration == 2:
        pathToParfile = restartParfile

        #This is neither a failure nor a success.
        if(pathToParfile == "none"):
          log.stp("Parfile specified as none for this test.")
          return runSucceeded

        if not runSucceeded:
          break
      #the masterDict entry is the parfile that will be used.
      parfileName = os.path.basename(pathToParfile)
      self.masterDict["parfile"] = parfileName

      # read the execution script from "exeScript"
      exeScriptFile = os.path.join(pathToFlashTest, pathToExeScript)
      if not os.path.isfile(exeScriptFile):
        log.err("File \"exeScript\" not found. Unable to run executable.\n" +
                "Skipping all runs.")
        return False


      lines = open(exeScriptFile).read().split("\n")
      lines = [line.strip() for line in lines
               if len(line.strip()) > 0 and not line.strip().startswith("#")]
      script = "\n".join(lines)
      self.masterDict["script"] = script
      script = self.masterDict["script"]  # do it this way so that any angle-bracket variables
                                        # in "exeScript" will be filled in by self.masterDict
      # determine 'pathToRunSummary'
      pathToRunSummary = os.path.join(pathToRunDir, "run_summary")

      # cd to output directory to run executable
      os.chdir(pathToRunDir)

      #plant the parfiles needed in the execution directory
      parfileText = open(pathToParfile).read()
      open(os.path.join(pathToRunDir, parfileName), 'w').write(parfileText)

      # obtain and record number of processors
      if not self.masterDict.has_key("numProcs"):
        self.masterDict["numProcs"] = 1
      open(pathToRunSummary,"a").write("numProcs: %s\n" % self.masterDict["numProcs"])

      # record mpirun invocation in "flash_call" file and in log
      open(os.path.join(pathToRunDir, "flash_call"), "w").write(script)
      log.stp(script)

      # get stdout/stderr and duration of execution and write to file
      out, err, duration, exitStatus = getProcessResults(script)

      open(os.path.join(pathToRunDir, "flash_output"),"a").write(out)
      if len(err) > 0:
        open(os.path.join(pathToRunDir, "flash_error"),"a").write(err)

      # record execution time in the run summary and logfile in human-readable form
      duration = secondsToHuman.convert(duration)
      open(pathToRunSummary,"a").write("wallClockTime: %s\n" % duration)
      log.info("duration of execution: %s" % duration)

      # search the parfile output directory for checkpoint files
      checkFiles = [f for f in os.listdir(pathToRunDir) if re.match(".*?_chk_\d+$", f)]

      # record number and names of checkpoint files in the run summary
      open(pathToRunSummary,"a").write("numCheckfiles: %s\n" % len(checkFiles))
      for checkFile in checkFiles:
        open(pathToRunSummary,"a").write("checkFile: %s\n" % checkFile)

      if iteration == 0:
        #this was a full run for restart transparency test, save final checkpoint delete all others
        if len(checkFiles) > 0:
          checkFiles.sort()
          os.rename(checkFiles.pop(), os.path.join(pathToRunDir, "chk_restart_longrun_9f402yfg"))
          for chk in checkFiles: os.remove(chk)
        else:
          log.stp("Long run did not produce expected checkpoints")
          if exitStatus == 0:
            log.stp(" - marking exit-status as failed")
            exitStatus = 3

      # An exit status of 0 means a normal termination without errors.
      if exitStatus == 0:
        log.stp("Process exit-status reports execution successful")
        runSucceeded = True
      else:
        log.stp("Process exit-status reports execution failed")
        pathToFlashExe = self.masterDict["pathToFlashExe"]  # set by FlashCompiler, above
        if len(pathToFlashExe) > 0 and pathToFlashExe[0] == '/':
          time.sleep(1)
          killallCmd = 'killall -v -HUP %s' % pathToFlashExe
#          log.stp('killallCmd is "%s"' % killallCmd)
          p = subprocess.Popen([killallCmd],shell=True,stderr=subprocess.PIPE,close_fds=True)
          killStatus = p.wait()
          if killStatus==0:
            out = p.stderr.read().strip()
            if len(out) > 0:
              log.stp(killallCmd)
              log.info(out)
#          else:
#            log.stp("Exit status %d from %s" % (killStatus, killallCmd))
        else:
          log.stp('pathToFlashExe is "%s"' % pathToFlashExe)
        runSucceeded = False
      # iteration loop
      iteration += 1

    # cd back to flashTest
    os.chdir(pathToFlashTest)

    return runSucceeded


#This is going to need an external file to keep last completed run.
#Use pre-existing sfocu machinery
class CompositeTester(ComparisonTester):

#   Def __init__(Self, Testobject):
#     """
#     We Must Check For A List Of Last Completed Comparison Tests, Hence
#     The Initialization Method Of The Composite Case Tester Needs
#     To Be Overridden To Check For This File.
#     """

#     Self.Owner = Testobject
#     Self.Masterdict = Testobject.Masterdict

     #Open A File In The Top Level Of The Flash Test Directory.
     #If Not There, Set A Sentinel For The Filename (Null)
#     Lastchangedfile = Open("Last_changed_checkpoint", 'R+')
#     For Lines In Lastchangedfile:



  def compareFiles(self, pathToFile1, pathToFile2, compOptions, compareYesterday=False):
    #return true if test is successful.  Return false if we encounter an error.

    #Let's break this out for readability sake.
    log = self.masterDict["log"] #from flashTest.py
    outfile = self.masterDict["outfile"] #from flashTest.py

    if not os.path.exists(pathToFile1):
      log.stp("\"%s\" does not exist." % pathToFile1)
      outfile.write("\"%s\" does not exist.\n" % pathToFile1)
      return False

    #pathToLocalArchive = self.masterDict["pathToLocalArchive"] #belongs to flashTest.py
    #pathToFile = os.path.join(pathToLocalArchive, shortPathToFile)
    pathToFlash = self.masterDict["pathToFlash"] #from flashTest.py

    pathToSfocu = self.masterDict.get("pathToSfocu", os.path.join(pathToFlash, "tools", "sfocu", "sfocu"))
    sfocuScript = self.masterDict.get("sfocuScript", pathToSfocu)

    cmdAndArgs = [sfocuScript]
    cmdAndArgs.append(compOptions)
    #presently only supporting tolerances should we allow for arbitrary flags?
    #if self.masterDict.has_key("sfocuFlags"):
     # cmdAndArgs.append.("%s" % self.masterDict["sfocuFlags"])

    if compareYesterday == False:
      if "errTol" in self.masterDict :
        cmdAndArgs.append("-e %s " % self.masterDict["errTol"])
      if "partErrTol" in self.masterDict :
        cmdAndArgs.append("-p %s " % self.masterDict["partErrTol"])

    locFile2 = pullfile(pathToFile2,log)
    if not locFile2:
      return False #something seriously failed with pullfile.

    cmd = "%s %s %s" % (" ".join(cmdAndArgs), pathToFile1, locFile2)
    #log.stp('SFOCU CMD: ' + cmd)
    #outfile.write('SFOCU CMD: ' + cmd)
    retval = self.compare(pathToFile1, locFile2, cmd)
    pushfile(pathToFile2)

    if not retval:
      return False #something catastrophically failed with sfocu.

    out, err, duration, exitStatus = retval

    if exitStatus == 0:
      log.stp("Process exit-status reports sfocu ran successfully.")
      outfile.write("<b>sfocu output:</b>\n" + out.strip() + "\n\n")

      if out.strip().endswith("SUCCESS"):
        log.stp("comparison of benchmark files yielded: SUCCESS")
        return True
      else:
        log.stp("comparison of benchmark files yielded: FAILURE")
        return False
    else:
      log.stp("Process exit-status reports sfocu encountered an error")
      #record what we got before the error.
      outfile.write("Process exit-status retports sfocu encountered an error\n" +
                    "<b>sfocu output:</b>\n" + out.strip() + "\n\n")
      return False
    return True

  def adjustFilesToDelete(self, checkpointChanged, restartChanged):

    """
    We can take files that we want to keep out of the files to delete file with
    this.  This file is written by the executer.
    """

    pathToRunDir = self.masterDict["pathToRunDir"]  #guaranteed to exist by flashTest.py
    checkpointFilename = None
    if checkpointChanged:
      filesToDelete = []


 #   deleteFile = open(os.path.join(pathToRunDir, "files_to_delete"), "r")

    #find the line to remove and do not include it
   # for line in deleteFile:
   #   if line == checkpointFilename:
   #     pass
   #   elif line == restartFilename:
   #     pass
   #   else:
   #     filesToDelete.append(line)

   #deleteFile.close()

   # open(os.path.join(pathToRunDir, "files_to_delete"), 'w').write("\n".join(filesToDelete))

    return

  def test(self):
    log                = self.masterDict["log"]           # guaranteed to exist by flashTest.py
    pathToFlash        = self.masterDict["pathToFlash"]   # guaranteed to exist by flashTest.py
    pathToRunDir       = self.masterDict["pathToRunDir"]  # guaranteed to exist by flashTest.py
    outfile            = self.masterDict["outfile"]       # guaranteed to exist by flashTest.py

    restartBenchmark = self.masterDict.get("restartBenchmark","")  # supposed to exist in "test.info"

    pathToSfocu = self.masterDict.get("pathToSfocu", os.path.join(pathToFlash, "tools", "sfocu", "sfocu"))
    sfocuScript = self.masterDict.get("sfocuScript", pathToSfocu)

    if "chkMax" not in self.masterDict:
      log.stp("No checkpoint files were produced, so no comparisons can be made.")
      outfile.write("No checkpoint files were produced, so no comparisons can be made.")
      return False
    else:
      chkMax = self.masterDict["chkMax"]  # should always have a value because even if execution produced
                                        # no checkpoint files, the seed-checkpoint copied into 'runDir'
                                        # should have been tagged as "chkMax" in ComparisonExecuter's
                                        # "adjustFilesToDelete" method.


    pathToChkMax = os.path.join(pathToRunDir, self.masterDict["chkMax"])

    #first, we do the last changed comparison. Located in localArchive/platform/test_name/test_par
    #Date of last changed in benchmark itself.  Append a file with changed on date.
    log.stp("Part 1: Compare this invocation's result to the last changed result.\n")
    outfile.write("Part 1: Compare this invocation's result to the last changed result.\n")

    #we have to make sure that the lastChanged file is

    #construct names
    targetFilename = self.masterDict["checkpointBasename"]
    targetFilename += self.masterDict["comparisonNumber"]
    pathToLocalArchive = self.masterDict["pathToLocalArchive"]
    pathToRunDir  =  self.masterDict["pathToRunDir"]
    pathToRunFile = os.path.join(pathToRunDir, targetFilename)

    siteDir = self.masterDict["siteDir"]
    buildDir = self.masterDict["buildDir"]
    runDir = self.masterDict["runDir"]

    pathToLastChanged = os.path.join(pathToLocalArchive, siteDir, buildDir, runDir, targetFilename)

    #ensure that the lastChanged file exists, if not, then we cannot run this comparison.
    if not os.path.isfile(pathToLastChanged):
      retval1 = False
      log.stp("ERROR: Unable to compare to the last changed checkpoint:\n %s not found.\n" % pathToLastChanged)
      outfile.write("ERROR: Unable to compare to the last changed checkpoint:\n %s not found.\n" % pathToLastChanged)
    else:
      pathToLastChangedDates = os.path.join(pathToLocalArchive, siteDir, buildDir, runDir, "lastChangedDates")
      out, err, duration, exitStatus = getProcessResults("tail -1 %s" % pathToLastChangedDates, 5, False)
      if (exitStatus == 0):
        log.stp("Last changed result is from %s.\n" % out.strip())
        outfile.write("Last changed result is from %s.\n" % out.strip())
      compFlags = " "
      if "sfocuFlags" in self.masterDict:
        compFlags = self.masterDict["sfocuFlags"]

      retval1 = self.compareFiles(pathToRunFile, pathToLastChanged, compFlags, compareYesterday=True)

    #self.masterDict["changedFromPrevious"] = False
    #no change, we can go on normally otherwise...
    #result changed! We have to do some movement and updating.  Unless this is a debug run.
    if retval1:
        self.masterDict["sameAsPrevious"] = True
    elif "usingDebugSetupper" in self.masterDict: #if a special debug run, should not archive
        self.masterDict["changedFromPrevious"] = True
    elif not os.path.exists(pathToRunFile):
        log.stp("not attempting to copy \"%s\".\n" % pathToRunFile)
        if os.path.isfile(pathToLastChanged):
          self.masterDict["changedFromPrevious"] = True
    else:
      #If this is the first time this test has been run, we need to generate
      #our supporting directories.
      if not os.path.isdir(os.path.join(pathToLocalArchive, siteDir,buildDir, runDir)):
        if not os.path.isdir(os.path.join(pathToLocalArchive, siteDir)):
          os.mkdir(os.path.join(pathToLocalArchive, siteDir))
        if not os.path.isdir(os.path.join(pathToLocalArchive, siteDir, buildDir)):
          os.mkdir(os.path.join(pathToLocalArchive, siteDir, buildDir))
        if not os.path.isdir(os.path.join(pathToLocalArchive, siteDir,buildDir, runDir)):
          os.mkdir(os.path.join(pathToLocalArchive, siteDir,buildDir, runDir))
      try:
        pathToLastChangedDates = os.path.join(pathToLocalArchive, siteDir, buildDir, runDir, "lastChangedDates")
        lastChangedDateFile = open(pathToLastChangedDates, 'a')

        log.stp("\nattempting copy\n")
        shutil.copy(pathToRunFile, pathToLastChanged)
        log.stp("copy good\n")
        #update the date of the file.
        lastChangedDateFile.write("%s\n" % self.masterDict["dateStr"]) #datetime.date.today())

        lastChangedDateFile.close()

        #set changedFromPrevious to let flashTest.py know that this has occured

        self.masterDict["changedFromPrevious"] = True
        log.stp("Today's run moved to the last changed benchmark.\n")
        outfile.write("Today's run moved to the last changed benchmark.\n")

      except IOError:
        log.err("Cannot open the last changed date file. File move aborted.\n")
        outfile.write("ERROR: Cannot open the last changed date file. File move aborted.\n")

    #Go on to benchmark test
    log.stp("Part 2: Compare this invocation's result to the approved comparison benchmark.\n")
    outfile.write("Part 2: Compare this invocation's result to the approved comparison benchmark.\n")

    def logmsg(msg):
      log.stp(msg + "\n")
      outfile.write(msg + "\n")
    def logerr(msg):
      log.err(msg + "\n")
      outfile.write(msg + "\n")

    self.masterDict.pushLayer()
    compFlags = " "       ################
    seededBenchmark = False
    if "comparisonBenchmark" not in self.masterDict:
      # no comparison benchmark, try and find one from the seed site if it exists
      pathSeedInfo = self.masterDict.get("benchmarkSeedInfo", None)
      if not pathSeedInfo:
        logerr("No comparisonBenchmark specified and benchmarkSeedInfo not set, so no comparisons being done.")
        return False
      locSeedInfo = pullfile(pathSeedInfo,log) # bring seed info file local if its remote
      if not locSeedInfo:
        logerr("Unable to retrieve remote seed Info file.")
        return False
      xmlSeed = parseXml(locSeedInfo) # parse it
      testPath = self.masterDict["testPath"]
      seedPath = os.path.join(self.masterDict["benchmarkSeedSite"], testPath)
      seedNode = xmlSeed.findChild(seedPath) # find the current path, but prepended with the seed site
      if seedNode is None:
        logmsg("Node '%s' not found in seed info file." % seedPath)
        return False
      seedDict = parser.parseLines(seedNode.text)

      ##==Get comparison benchmark from remote host===

      if "comparisonBenchmark" not in seedDict:
        logerr("No benchmark declared in seed info file for comparisons. No comparisons being done.")
        return False
      else:
        logmsg("Comparison benchmark being ported from mainArchive and/or seed platform.")

      self.masterDict.pushLayer()
      self.masterDict.update({'siteDir':'<benchmarkSeedSite>'})
      comparisonBenchmark = self.masterDict.dereferencePointers(seedDict["comparisonBenchmark"])
      self.masterDict.popLayer()

      pathToBenchmark = os.path.join(self.masterDict["benchmarkSeedArchive"], comparisonBenchmark)

      if "errTol" not in self.masterDict:
        self.masterDict["errTol"] = str(self.masterDict.get("benchmarkSeedErrTol", "1.e-12"))
      if "partErrTol" not in self.masterDict:
        self.masterDict["partErrTol"] = str(self.masterDict.get("benchmarkSeedPartErrTol", "1.e-12"))
      seededBenchmark = True
    else:
      comparisonBenchmark = self.masterDict["comparisonBenchmark"]
      pathToBenchmark = os.path.join(pathToLocalArchive, comparisonBenchmark)
      if "sfocuFlags" in self.masterDict:
        compFlags = self.masterDict["sfocuFlags"]

    targetFilename = self.masterDict["checkpointBasename"] + self.masterDict["comparisonNumber"]
    pathToRunDir = self.masterDict["pathToRunDir"]
    pathToRunFile = os.path.join(pathToRunDir, targetFilename)

    retval2 = self.compareFiles(pathToRunFile, pathToBenchmark, compFlags)
    self.masterDict.popLayer()

    #if seededBenchmark:
    #  pushfile(remotePathToBenchmark) #this will delete locally pulled file
    if seededBenchmark and retval2:
      logmsg("Comparison of seeded benchmark was within tolerance, approving benchmark...")
      self.masterDict["testXmlNode"].text.append("comparisonBenchmark: <siteDir>/%s/<buildDir>/<runDir>/<checkpointBasename><comparisonNumber>" % self.masterDict["dateStr"])
      self.masterDict["testXmlNode"].smudge()

    #Test the restart against a restart benchmark.
    logmsg("Part 3: Compare this invocation's result to the approved restart benchmark.")
    retval3 = False

    targetFilename = self.masterDict["checkpointBasename"] + self.masterDict["restartNumber"]
    pathToRestartFile = os.path.join(pathToRunDir, targetFilename)

    if "restartBenchmark" not in self.masterDict:
      # transparency should have been tested and produced chk_restart_longrun_9f402yfg
      if retval2: # did comparison benchmark match?
        logmsg("Restart benchmark not present, testing for a transparent restart...")
        pathToLongChk = os.path.join(pathToRunDir, "chk_restart_longrun_9f402yfg")
        retval3 = self.compareFiles(pathToLongChk, pathToRestartFile, "")
        if retval3:
          logmsg("Restart transparency confirmed, approving restart benchmark in test xml file.")
          self.masterDict["testXmlNode"].text.append("restartBenchmark: <siteDir>/%s/<buildDir>/<runDir>/<checkpointBasename><restartNumber>" % self.masterDict["dateStr"])
          self.masterDict["testXmlNode"].smudge()
        else:
          logmsg("Restart was not transparent, not approving restart benchmark.")
      else:
        logerr("No restart benchmark specified and previous comparison failed, taking no action.")
    elif "restartNumber" not in self.masterDict:
      logerr("restartBenchmark not found in test.info for this test!")
    else:
      restartBenchmark = self.masterDict["restartBenchmark"]
      pathToRestartBenchmark = os.path.join(pathToLocalArchive, restartBenchmark)

      #make sure that the restart checkpoint exists locally
      if not os.path.isfile(pathToRestartFile):
        logerr("The restart test checkpoint file from this run was not found! %s" % pathToRestartFile)
      else:
        compFlags = " "
        if "sfocuFlags" in self.masterDict:
          compFlags = self.masterDict["sfocuFlags"]

        retval3 = self.compareFiles(pathToRestartFile, pathToRestartBenchmark, compFlags)

    return retval2 and retval3

#new tester for Chris's experimental IO tool.
#DEV: Stopgap until we get things properly intergrated into SFOCU

class CompositeTesterExperimentalIO(CompositeTester):

    #This is all we should need to override. -PR
  def compareFiles(self, pathToFile1, pathToFile2, compOptions):
    #return true if test is successful.  Return false if we encounter an error.

    #Let's break this out for readability sake.
    log = self.masterDict["log"] #from flashTest.py
    outfile = self.masterDict["outfile"] #from flashTest.py
    pathToFlash = self.masterDict["pathToFlash"] #from flashTest.py

    pathToCmpExpIO = self.masterDict.get("pathToCmpExpIO", os.path.join(pathToFlash, "tools", "cmpExpIO", "cmpExpIO"))
    cmpExpIOScript = self.masterDict.get("cmpExpIOScript", "cmpExpIO")

    cmdAndArgs = [cmpExpIOScript]
    cmdAndArgs.append(compOptions)
    # This tool takes no flags at the mo

    cmd = "%s %s %s" % (" ".join(cmdAndArgs), pathToFile1, pathToFile2)
    retval = self.compare(pathToFile1, pathToFile2, cmd)

    if not retval:
      return False #something catastrophically failed with sfocu.

    out, err, duration, exitStatus = retval

    if exitStatus == 0:
      log.stp("Process exit-status reports CmpExpIO ran successfully.")
      outfile.write("<b>CmpExpIO output:</b>\n" + out.strip() + "\n\n")

      if out.strip().endswith("SUCCESS"):
        log.stp("comparison of benchmark files yielded: SUCCESS")
        return True
      else:
        log.stp("comparison of benchmark files yielded: FAILURE")
        return False
    else:
      log.stp("Process exit-status reports CmpExpIO encountered an error")
      #record what we got before the error.
      outfile.write("Process exit-status retports CmpExpIO encountered an error\n" +
                    "<b>CmpExpIO output:</b>\n" + out.strip() + "\n\n")
      return False
    return True



import re
import six
import sys, os

sys.path.insert(0,os.path.dirname(os.path.abspath(__file__)))

class LayeredDict:
  """
  Object that contains a list of dictionaries, 'dicts', and implements
  many of the methods of a python3 dictionary.

  The get, set, and del operations:

    foo = d["foo"]

    d["foo"] = foo

    del d["foo"]

  operate on the dictionary referenced by the value of 'activeLayer'.

  The 'setActiveLayer' method sets this value and deletes the dictionary
  at the specified value as well as all dictionaries at higher-indexed
  positions. See the docstring for setActiveLayer() for more information.

  Additionally, LayeredDict allows keys to have values that are pointers
  to other keys. See the docstring for the "__getitem__" method for more
  information.
  """
  pat1 = re.compile("(<(.*?)>)")  # regex used to look for pointers in values

  def __init__(self, d={}):
    self.activeLayer = 0
    self.dicts = []
    if d.__class__.__name__ == "dict":
      self.dicts.append(d)
    elif d.__class__.__name__ == "LayeredDict":
      for innerDict in d.dicts:
        self.dicts.append(innerDict.copy())
    else:
      dType = type(d).__name__
      raise TypeError("Objects of type '%s' may not be passed to the contructor of class LayeredDict." % dType)

  def __contains__(self, key):
    if self.has_key(key):
      return True
    return False

  def __delitem__(self, key):
    return self.dicts[self.activeLayer].__delitem__(key)

  def __getitem__(self, key):
    """
    Examine the value of the given key and resolve any pointers,
    which are strings enclosed in angle brackets, to real values
    from the dictionary, if available. Return the new value.

    e.g., if LayeredDict 'd' has key/value pairs:

      foo: 'Mike'
      bar: '<foo>'

    Querying d["bar"] yields 'Mike'

    The presence of multiple pointer elements and/or non-pointer
    does not cause a problem.

    e.g., if LayeredDict 'd' has key/value pairs:

      foo: 'Mike'
      bar: '"My name is <foo>", said <foo>'

    Querying d["bar"] yields '"My name is Mike", said Mike'

    A pointer whose value is keyed to other pointers will still
    resolve to a real value as long as one is available, regardless
    of the length of the list of references in between.

    e.g., if LayeredDict 'd' has key/value pairs:

      foo: 'Mike'
      bar: '<foo>'
      baz: '<bar>'

    Querying d["baz"] still yields "Mike"

    A string enclosed in angle-brackets that does not correspond to
    a key in the dictionary will be returned normally.

    e.g., if LayeredDict 'd' has key/value pairs:

      foo: 'Mike'
      bar: '<foo>'
      baz: '<bupkus>'

    Querying d["baz"] yields the string "<bupkus>"

    Pointers contained inside lists, tuples, and dictionaries will be
    resolved wherever possible. This extends to recursively nested
    lists, tuples, dictionaries, etc.

    Note, however, that a pointer may only be resolved to a key whose
    corresponding value is a string or another pointer or sequence of
    pointers that ultimately terminate in a string.

    e.g., if LayeredDict 'd' has key/value pairs:

      foo: 'Mike'
      bar: ['hello', '<foo>']
      baz: '<bar>'

    Querying d["bar"] yields the list ['hello', 'Mike']
    but querying d["baz"] causes a TypeError.

    At first it might seem that d["baz"] should return the same as
    d["bar"], but consider the case where "baz" is keyed to a string
    that contains more elements than just the pointer...

      baz: 'Here is a list: <bar>'

    In this case it is not clear exactly how the referent of "<bar>",
    a list, should be joined to the string elements of "baz". This
    use case therefore raises the TypeError.
    """
    n = self.activeLayer
    while n >= 0:
      if key in self.dicts[n]:
        val = self.dicts[n][key]
        return self.dereferencePointers(val)
      else:
        n -= 1
    raise KeyError

  def __repr__(self):
    return self.__str__()

  def __setitem__(self, key, value):
    return self.dicts[self.activeLayer].__setitem__(key, value)

  def __str__(self):
    s = ""
    for i in range(len(self.dicts)):
      s += "%s: %s\n" % (i, self.dicts[i])
    return s.strip()

  def __dereferencePointers(self, val):
    """
    Recursively search through lists, tuples, and dictionaries
    for pointers (strings), dereferencing wherever possible.
    """
    if six.PY2:
      if isinstance(val, unicode):
        val = val.encode('utf8')     
        soughtOpts = self.pat1.findall(val)
        for soughtOpt, soughtOptStripped in soughtOpts:
          if self.has_key(soughtOptStripped):
            val = val.replace(soughtOpt, self[soughtOptStripped])
          else:
            print (val)  
            val = ''
    if isinstance(val, str):    
      soughtOpts = self.pat1.findall(val)
      for soughtOpt, soughtOptStripped in soughtOpts:
        if self.has_key(soughtOptStripped):
          val = val.replace(soughtOpt, self[soughtOptStripped])
        else:  
          print(val)
          val = ''        
    elif isinstance(val, list):
      val = [self.__dereferencePointers(item) for item in val]
    elif isinstance(val, tuple):
      val = tuple([self.__dereferencePointers(item) for item in val])
    elif isinstance(val, dict):
      val = dict([(key, self.__dereferencePointers(val[key])) for key in val])
    return val

  def dereferencePointers(self, val):
    return self.__dereferencePointers(val)

  def copy(self):
    return LayeredDict(self)

  def despecify(self, keyword=""):
    for key in self.keys():
      m = re.match("(%s\.)(.*)" % keyword, key)
      if m: self[m.group(2)] = self[key]

  def get(self, key, alt=None):
    if self.has_key(key):
      return self[key]
    return alt

  def has_key(self, key):
    for d in self.dicts:
      if key in d:
        return True
    return False

  def keys(self):
    keysDict = {}
    for d in self.dicts:
      for key in d.keys():
        keysDict[key] = None
    return keysDict.keys()

  def setActiveLayer(self, n):
    if n < len(self.dicts):
      self.activeLayer = n
      while n < len(self.dicts):
        del self.dicts[n]
    else:
      self.activeLayer = len(self.dicts)
    self.dicts.append({})

  def pushLayer(self):
    self.activeLayer = len(self.dicts)
    self.dicts.append({})

  def popLayer(self):
    self.activeLayer -= 1
    self.dicts.pop()

  def update(self, d):
    return self.dicts[self.activeLayer].update(d)



import tempfile
import subprocess
import time
import uuid
import os,sys

sys.path.insert(0,os.path.dirname(os.path.abspath(__file__)))

def _run(args, input=None):
  p = subprocess.Popen(args, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
  stdout, stderr = p.communicate(input)
  if p.returncode != 0: raise Exception(subprocess.list2cmdline(args) + ':\n' + stderr)
  return stdout

def splitHostAndPath(path):
  if path.count(':') > 0:
    x = path.split(':', 1)
    return (x[0] or 'localhost', x[1])
  else:
    return ('localhost', path)

# returns contents of file as string
def readfile(path):
  host, path = splitHostAndPath(path)
  if host != 'localhost':
    tmp = tempfile.NamedTemporaryFile(delete=False)
    tmp.close()
    tmp = tmp.name
    try:
      _run(['scp','%s:%s'%(host,path),tmp])
      with open(tmp) as f:
        data = f.read()
    finally:
      os.remove(tmp)
  else:
    with open(path) as f:
      data = f.read()
  return data

# writes contents of file given string
def writefile(path, data):
  host, path = splitHostAndPath(path)
  if host != 'localhost':
    w = tempfile.NamedTemporaryFile(delete=False)
    w.write(data)
    w.close()
    w = w.name
    try:
      _run(['scp',w,'%s:%s'%(host,path)])
    finally:
      os.remove(w)
  else:
    with open(path,'w') as w:
      w.write(data)

# delta: string -> string
def updatefile(path, delta):
  def mv(src,dst): # move (hopefully atomically) a file from src to dst
    shost, spath = splitHostAndPath(src)
    dhost, dpath = splitHostAndPath(dst)
    if shost != dhost: raise ValueError('cannot move files on different hosts')
    if shost != 'localhost':
      _run(['ssh', shost, subprocess.list2cmdline(['mv',spath,dpath])])
    else:
      os.rename(spath, dpath)
  held = path + ".lock-" + uuid.uuid4().hex
  attempts = 0
  max_attempts = 5
  while attempts < max_attempts:
    try:
      mv(path, held)
      break
    except Exception as e:
      attempts += 1
      if attempts == max_attempts:
        raise Exception('Could not obtain lock on file %s:\n%s' % (path,str(e)))
      time.sleep(attempts) # wait 'attempts' number of seconds
  try: # we've got the file on hold: read it, change it, write it
    data = readfile(held)
    data = delta(data)
    writefile(held, data)
  finally:
    try: # and release the lock
      mv(held, path)
    except Exception as e:
      raise Exception('Could not release lock on file %s:\n' % (held,str(e)))




def convert(seconds):
  """
  Takes a number of seconds and returns a human-readable conversion.
  
  Input of s <= 59 seconds will yield a string of the form "s seconds".
  Higher values will yield a string of the form "mm:ss", "hh:mm:ss",
  or "dd:hh:mm:ss" - whichever is appropriate for the given input.
  """

  # round seconds to nearest second
  iseconds = int(round(seconds))

  SECONDS_IN_DAY = 86400
  SECONDS_IN_HOUR = 3600
  SECONDS_IN_MINUTE = 60

  daysStr    = ""
  hoursStr   = ""
  minutesStr = ""
  secondsStr = ""

  days = iseconds // SECONDS_IN_DAY
  if (days):
    daysStr = ("%2d" % days).replace(" ","0")
    iseconds -= (days * SECONDS_IN_DAY)

  hours = iseconds // SECONDS_IN_HOUR
  if (hours or days):
    hoursStr = ("%2d" % hours).replace(" ","0")
    iseconds -= (hours * SECONDS_IN_HOUR)

  minutes = iseconds // SECONDS_IN_MINUTE
  if (minutes or hours or days):
    minutesStr = ("%2d" % minutes).replace(" ","0")
    iseconds -= (minutes * SECONDS_IN_MINUTE)

  if (minutes or hours or days):
    secondsStr = ("%2d" % iseconds).replace(" ","0")
    l = [daysStr, hoursStr, minutesStr, secondsStr]
    conversionStr = ":".join([str for str in l if len(str) > 0])
  elif seconds == 1:
    conversionStr = "1 second"
  elif seconds < 9.5:
    conversionStr = "%8.6f seconds" % seconds
  else:
    conversionStr = "%s seconds" % iseconds

  return conversionStr
